from autograd import numpy as np

class InverseExponential_():
    def __init__(self):
        self.name = 'InverseExponential'
        self.phi_param_map = {'a' : 0, 'b' : 1}
        self.phi_bounds = ((None, None), (0, None),)

    def phi(self, X, *params):
        a = params[0]
        b = params[1]
        return 1./(b * np.exp(a/X))

    def phi_init(self, life, X):
        X = X.flatten()
        a, b = (np.polyfit(1./X, np.log(1./life), 1))
        return [a, np.exp(b)]

InverseExponential = InverseExponential_()

class Exponential_():
    def __init__(self):
        self.name = 'Exponential'
        self.phi_param_map = {'a' : 0, 'b' : 1}
        self.phi_bounds = ((None, None), (0, None),)

    def phi(self, X, *params):
        a = params[0]
        b = params[1]
        return b * np.exp(a/X)

    def phi_init(self, life, X):
        X = X.flatten()
        a, b = (np.polyfit(1./X, np.log(life), 1))
        return [a, np.exp(b)]

Exponential = Exponential_()


from scipy.optimize import minimize
from surpyval import np
from autograd import jacobian, hessian
import sys


def mps_fun(params, dist, x, inv_trans, const, c, n, offset):
    if offset:
        x = x - inv_trans(const(params))[0]
        params = inv_trans(const(params))[1:]
    else:
        params = inv_trans(const(params))
    return dist.neg_mean_D(x, c, n, *params)


def mps(model):
    """
    MPS: Maximum Product Spacing

    This is the method to get the largest (geometric) average distance
    between all points. This method works really well when all points are
    unique. Some complication comes in when using repeated data. This method
    is quite good for offset distributions.
    """

    old_err_state = np.seterr(all='ignore')

    dist = model.dist
    x, c, n = model.data['x'], model.data['c'], model.data['n']
    const = model.fitting_info['const']
    inv_trans = model.fitting_info['inv_trans']
    init = model.fitting_info['init']
    offset = model.offset

    jac = jacobian(mps_fun)
    hess = hessian(mps_fun)

    res = minimize(mps_fun, init,
                   method='Newton-CG',
                   jac=jac,
                   hess=hess,
                   tol=1e-15,
                   args=(dist, x, inv_trans, const, c, n, offset))

    if (res.success is False) or (np.isnan(res.x).any()):
        res = minimize(mps_fun, init, method='BFGS', jac=jac,
                       args=(dist, x, inv_trans, const, c, n, offset))

    if (res.success is False) or (np.isnan(res.x).any()):
        res = minimize(mps_fun, init,
                       args=(dist, x, inv_trans, const, c, n, offset))

    if (res.success is False) or (np.isnan(res.x).any()):
        print("MPS FAILED: Try alternate estimation method", file=sys.stderr)

    results = {}
    params = inv_trans(const(res.x))
    results['res'] = res

    if offset:
        results['gamma'] = params[0]
        results['params'] = params[1::]
    else:
        results['params'] = params

    results['jac'] = jac

    np.seterr(**old_err_state)

    return results



from surpyval import np
from scipy.stats import pearsonr
from surpyval.nonparametric import plotting_positions
from scipy.optimize import minimize


def mpp(model):
    """
    MPP: Method of Probability Plotting

    This is the classic probability plotting paper method. This method
    creates the plotting points, transforms it to Weibull scale and then fits
    the line of best fit.
    """
    dist = model.dist
    x, c, n, t = (model.data['x'], model.data['c'],
                  model.data['n'], model.data['t'])

    heuristic = model.fitting_info['heuristic']
    on_d_is_0 = model.fitting_info['on_d_is_0']
    offset = model.offset
    rr = model.fitting_info['rr']
    turnbull_estimator = model.fitting_info['turnbull_estimator']

    if rr not in ['x', 'y']:
        raise ValueError("rr must be either 'x' or 'y'")

    if hasattr(dist, 'mpp'):
        return dist.mpp(x, c, n, heuristic=heuristic, rr=rr,
                        on_d_is_0=on_d_is_0, offset=offset)

    x_, r, d, F = plotting_positions(x=x, c=c, n=n, t=t,
                                     heuristic=heuristic,
                                     turnbull_estimator=turnbull_estimator)

    x_mask = np.isfinite(x_)
    x_ = x_[x_mask]
    F = F[x_mask]
    d = d[x_mask]
    r = r[x_mask]

    if not on_d_is_0:
        x_ = x_[d > 0]
        y_ = F[d > 0]
    else:
        y_ = F

    mask = (y_ != 0) & (y_ != 1)
    y_pp = y_[mask]
    x_pp = x_[mask]
    y_pp = dist.mpp_y_transform(y_pp)

    if offset:
        # I think this should be x[c != 1] and not in xl
        x_min = np.min(x_pp)

        # fun = lambda gamma : -pearsonr(np.log(x - gamma), y_)[0]
        def fun(gamma):
            g = x_min - np.exp(-gamma)
            out = -pearsonr(dist.mpp_x_transform(x_pp - g), y_pp)[0]
            return out

        res = minimize(fun, 0.)
        gamma = x_min - np.exp(-res.x[0])
        x_pp = x_pp - gamma

    x_pp = dist.mpp_x_transform(x_pp)

    if rr == 'y':
        params = np.polyfit(x_pp, y_pp, 1)
    elif rr == 'x':
        params = np.polyfit(y_pp, x_pp, 1)

    params = dist.unpack_rr(params, rr)

    results = {}

    if offset:
        results['gamma'] = gamma

    results['params'] = params
    results['rr'] = rr

    return results



import autograd.numpy as np
from autograd import jacobian, hessian, elementwise_grad
from scipy.optimize import minimize
import surpyval
import inspect
import warnings

from ..parametric.fitters import bounds_convert, fix_idx_and_function
from .regression import Regression

class ParameterSubstitutionFitter():
    def __init__(self, kind, name, distribution, life_model, life_parameter, baseline=[],
                 param_transform=None, inverse_param_transform=None):

        if str(inspect.signature(life_model.phi)) != '(X, *params)':
            raise ValueError('PH function must have the signature \'(X, *params)\'')

        if type(baseline) != list:
            # Baseline used if using a function that deviates from some number, e.g. np.exp(np.dot(X, beta))
            baseline = [baseline]

        self.name = name
        self.kind = kind
        self.dist = distribution
        self.life_model = life_model
        self.k_dist = len(self.dist.param_names)
        self.bounds = self.dist.bounds
        self.support = self.dist.support
        self.param_names = self.dist.param_names
        self.param_map = {v : i for i, v in enumerate(self.dist.param_names)}
        self.phi = life_model.phi
        self.Hf_dist = self.dist.Hf
        self.hf_dist = self.dist.hf
        self.sf_dist = self.dist.sf
        self.ff_dist = self.dist.ff
        self.df_dist = self.dist.df
        self.baseline = baseline
        self.life_parameter = life_parameter
        self.fixed = {life_parameter : 1.}

        if param_transform is None:
            self.param_transform = lambda x: x
            self.inverse_param_transform = lambda x: x
        else:
            self.param_transform = param_transform
            self.inverse_param_transform = inverse_param_transform

    def Hf(self, x, X, *params):
        x = np.array(x)
        if np.isscalar(X):
            X = np.ones_like(x) * X
        else:
            X = np.array(X)

        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])

        Hf = np.zeros_like(x)
        stresses = np.unique(X, axis=0)
        for stress in stresses:
            life_param_mask = np.array(range(0, len(dist_params))) == self.param_map[self.life_parameter]
            dist_params_i = np.where(life_param_mask,
                self.param_transform(self.phi(stress, *phi_params)),
                dist_params)
            mask = (X == stress).all(axis=1)
            Hf = np.where(mask, self.Hf_dist(x, *dist_params_i), Hf)

        return Hf

    def hf(self, x, X, *params):
        x = np.array(x)
        if np.isscalar(X):
            X = np.ones_like(x) * X
        else:
            X = np.array(X)

        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])

        hf = np.zeros_like(x)
        for stress in np.unique(X, axis=0):
            life_param_mask = np.array(range(0, len(dist_params))) == self.param_map[self.life_parameter]
            params = np.where(life_param_mask,
                self.param_transform(self.phi(stress, *phi_params)),
                dist_params)
            mask = (X == stress).all(axis=1)
            hf = np.where(mask, self.hf_dist(x, *params), hf)

        return hf

    def df(self, x, X, *params):
        x = np.array(x)
        if np.isscalar(X):
            X = np.ones_like(x) * X
        else:
            X = np.array(X)
        return self.hf(x, X, *params) * np.exp(-self.Hf(x, X, *params))

    def sf(self, x, X, *params):
        x = np.array(x)
        if np.isscalar(X):
            X = np.ones_like(x) * X
        else:
            X = np.array(X)
        return np.exp(-self.Hf(x, X, *params))

    def ff(self, x, X, *params):
        x = np.array(x)
        if np.isscalar(X):
            X = np.ones_like(x) * X
        else:
            X = np.array(X)
        return 1 - np.exp(-self.Hf(x, X, *params))

    def _parameter_initialiser_dist(self, x, c=None, n=None, t=None):
        out = []
        for low, high in self.bounds:
            if (low is None) & (high is None):
                out.append(0)
            elif high is None:
                out.append(low + 1.)
            elif low is None:
                out.append(high - 1.)
            else:
                out.append((high + low)/2.)

        return out

    def mpp_inv_y_transform(self, y, *params):
        return y

    def mpp_y_transform(self, y, *params):
        return y

    def mpp_inv_y_transform(self, y, *params):
        return y

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma

    def log_df(self, x, X, *params):
        return np.log(self.hf(x, X, *params)) - self.Hf(x, X, *params)

    def log_sf(self, x, X, *params):
        return -self.Hf(x, X, *params)

    def log_ff(self, x, X, *params):
        return np.log(self.ff(x, X, *params))

    def random(self, size, X, *params):
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])

        x = []
        X_out = []
        if type(X) == tuple:
            X = np.random.uniform(*X, size)

        for stress in np.unique(X, axis=0):
            life_param_mask = np.array(range(0, len(dist_params))) == self.param_map[self.life_parameter]
            dist_params = np.where(life_param_mask,
                                   self.param_transform(self.phi(stress, *phi_params)),
                                   dist_params)

            U = np.random.uniform(0, 1, size)
            x.append(self.dist.qf(U, *dist_params))
            if np.isscalar(stress):
                cols = 1
            else:
                cols = len(stress)
            X_out.append((np.ones((size, cols)) * stress))
        return np.array(x).flatten(), np.concatenate(X_out)

    def neg_ll(self, X, x, c, n, *params, verbose=False):
        like = np.zeros_like(x).astype(float)
        like = np.where(c ==  0, self.log_df(x, X, *params), like)
        like = np.where(c ==  1, self.log_sf(x, X, *params), like)
        like = np.where(c ==  -1, self.log_ff(x, X, *params), like)
        like = np.multiply(n, like)
        like = -np.sum(like)
        if verbose:
            print('   ' * 100, end='\r')
            print(like, ": ", params, end='\r')
        return like

    def fit(self, X, x, c=None, n=None, t=None, init=[], fixed={}, verbose=False):
        x, c, n, t = surpyval.xcnt_handler(x=x, c=c, n=n, t=t, group_and_sort=False)

        if init == []:
            stress_data = []
            params_at_X = []

            # How do I make this work when there is only one failure per stress?
            for s in np.unique(X, axis=0):
                mask = (X == s).all(axis=1)
                with warnings.catch_warnings():
                    warnings.filterwarnings('error')
                    try:
                        params_at_X.append(self.dist.fit(x[mask], c[mask], n[mask]).params)
                        stress_data.append(s)
                    except np.RankWarning:
                        pass
                    finally:
                        pass

            params_at_X = np.array(params_at_X)
            stress_data = np.array(stress_data)
            dist_init = params_at_X.mean(axis=0)

            i = self.param_map[self.life_parameter]

            if len(params_at_X) < 2:
                raise ValueError("Insufficient data at separate X values. Try manually setting initial guess using `init` keyword in `fit`")

            parameter_data = params_at_X[:, i]

            parameter_data = self.inverse_param_transform(parameter_data)

            if callable(self.life_model.phi_init):
                if str(inspect.signature(self.life_model.phi_init)) == '(X)':
                    phi_init = self.life_model.phi_init(X)
                else:
                    phi_init = self.life_model.phi_init(parameter_data, stress_data)
            else:
                phi_init = self.life_model.phi_init

            init = np.array([*dist_init, *phi_init])
        else:
            init = np.array(init)

        if self.baseline != []:
            baseline_model = self.dist.fit(x, c, n, t)
            baseline_fixed = {k : baseline_model.params[baseline_model.param_map[k]] for k in self.baseline}
            fixed = {**baseline_fixed, **fixed}

        if self.fixed != {}:
            fixed = {**self.fixed, **fixed}

        # Dynamic or static bounds determination
        if callable(self.life_model.phi_bounds):
            bounds = (*self.bounds, *self.life_model.phi_bounds(X))
        else:
            bounds = (*self.bounds, *self.life_model.phi_bounds)

        if callable(self.life_model.phi_param_map):
            phi_param_map = self.life_model.phi_param_map(X)
        else:
            phi_param_map = self.life_model.phi_param_map

        param_map = {**self.param_map, **{k : v + len(self.param_map) for k, v in phi_param_map.items()}}
        self.param_map = param_map

        transform, inv_trans, funcs, inv_f = bounds_convert(x, bounds)
        const, fixed_idx, not_fixed = fix_idx_and_function(fixed, param_map, funcs)

        init = transform(init)[not_fixed]

        with np.errstate(all='ignore'):
            fun  = lambda params, verbose : self.neg_ll(X, x, c, n, *inv_trans(const(params)), verbose=verbose)
            # jac = jacobian(fun)
            # hess = hessian(fun)
            res = minimize(fun, init, args=(verbose))
            res = minimize(fun, res.x, args=(verbose), method='TNC')
            # res = minimize(fun, init, jac=jac, method='BFGS')
            # res = minimize(fun, init, method='Newton-CG', jac=jac)

        params = inv_trans(const(res.x))
        dist_params = np.array(params[0:self.k_dist])
        phi_params = np.array(params[self.k_dist:])

        model = Regression()
        model.model = self
        model.kind = self.kind
        model.distribution = self.dist
        model.reg_model = self.life_model
        model.params = np.array(params)
        model.dist_params = dist_params
        model.phi_params = phi_params
        model.res = res
        model._neg_ll = res['fun']
        model.fixed = self.fixed
        model.k_dist = self.k_dist

        model.k = len(bounds)

        model.data = {
            'x' : x,
            'c' : c,
            'n' : n,
            't' : t
        }

        return model



import numpy as np
from surpyval import nonparametric as nonp
from surpyval import xcnt_handler


def filliben(x, c, n, t):
    """
    Method From:
    Filliben, J. J. (February 1975),
    "The Probability Plot Correlation Coefficient Test for Normality",
    Technometrics, American Society for Quality, 17 (1): 111-117
    """
    x, c, n, t = xcnt_handler(x, c, n, t)

    x = np.repeat(x, n)
    c = np.repeat(c, n)
    n = np.ones_like(x)

    idx = np.argsort(c, kind='stable')
    x = x[idx]
    c = c[idx]

    idx2 = np.argsort(x, kind='stable')
    x = x[idx2]
    c = c[idx2]
    N = len(x)

    ranks = nonp.rank_adjust(x, c)
    d = 1 - c
    r = np.linspace(N, 1, num=N)

    F = (ranks - 0.3175) / (N + 0.365)
    F[0] = 1 - (0.5 ** (1. / N))
    F[-1] = 0.5 ** (1. / N)

    out = {k: v for k, v in zip(['x', 'r', 'd'], (x, r, d))}
    out['R'] = 1 - F
    return out



from surpyval import np
from scipy.integrate import quad

from ..nonparametric import plotting_positions as pp

import surpyval
from surpyval import parametric as para

import pandas as pd
from copy import copy

from .fitters.mom import mom
from .fitters.mle import mle
from .fitters.mps import mps
from .fitters.mse import mse
from .fitters.mpp import mpp

from .fitters import bounds_convert, fix_idx_and_function

PARA_METHODS = ['MPP', 'MLE', 'MPS', 'MSE', 'MOM']
METHOD_FUNC_DICT = {
    'MPP': mpp,
    'MOM': mom,
    'MLE': mle,
    'MPS': mps,
    'MSE': mse
}


def init_from_bounds(dist):
    out = []
    for low, high in dist.bounds:
        if (low is None) & (high is None):
            out.append(0)
        elif high is None:
            out.append(low + 1.)
        elif low is None:
            out.append(high - 1.)
        else:
            out.append((high + low)/2.)

    return out


class ParametricFitter():
    def log_df(self, x, *params):
        return np.log(self.hf(x, *params)) - self.Hf(x, *params)

    def log_sf(self, x, *params):
        return -self.Hf(x, *params)

    def log_ff(self, x, *params):
        return np.log(self.ff(x, *params))

    def like(self, x, c, n, *params):
        like = np.zeros_like(x).astype(float)
        like = np.where(c == 0, self.df(x, *params), like)
        like = np.where(c == -1, self.ff(x, *params), like)
        like = np.where(c == 1, self.sf(x, *params), like)
        return like

    def log_like(self, x, c, n, p, f0, *params):
        # This is getting a bit much....
        like = np.zeros_like(x).astype(float)
        if f0 == 0:
            like = np.where(c == 0, np.log(p) + self.log_df(x, *params), like)
            like = np.where(c == -1, np.log(p) + self.log_ff(x, *params), like)
            if p == 1:
                like = np.where(c == 1, self.log_sf(x, *params), like)
            else:
                like = np.where(c == 1,
                                np.log(1 - (p * self.ff(x, *params))),
                                like)
        else:
            like = np.where(c == 0,
                            np.log(1 - f0) + np.log(p) +
                            self.log_df(x, *params),
                            like)
            like = np.where(c == -1,
                            np.log(1 - f0) + np.log(p) +
                            self.log_ff(x, *params),
                            like)
            like = np.where((c == 0) & (x == 0), np.log(f0), like)
            if p == 1:
                like = np.where(c == 1,
                                np.log(1 - f0) + self.log_sf(x, *params),
                                like)
            else:
                like = np.where(c == 1,
                                np.log(1 - f0) +
                                np.log(1 - (p * self.ff(x, *params))),
                                like)

        return like

    def log_like_i(self, x, c, n, inf_c_flags, p, f0, *params):
        ir = np.where(inf_c_flags[:, 1] == 1,
                      1,
                      (1 - f0) * p * self.ff(x[:, 1], *params))
        il = np.where(inf_c_flags[:, 0] == 1,
                      0,
                      (1 - f0) * p * self.ff(x[:, 0], *params))
        like_i = ir - il
        like_i = np.where(c != 2, 1., like_i)
        return np.log(like_i)

    def like_i(self, x, c, n, inf_c_flags, *params):
        # This makes sure that any intervals that are at the boundaries
        # of support or are infinite will not cause the autograd functions
        # to fail.
        ir = np.where(inf_c_flags[:, 1] == 1, 1, self.ff(x[:, 1], *params))
        il = np.where(inf_c_flags[:, 0] == 1, 0, self.ff(x[:, 0], *params))
        like_i = ir - il
        return like_i

    def like_t(self, t, t_flags, *params):
        # Needs to be updated to work with zi and ds models.
        # until then, can prevent it working in the `fit` method
        tr_denom = np.where(t_flags[:, 1] == 1, self.ff(t[:, 1], *params), 1.)
        tl_denom = np.where(t_flags[:, 0] == 1, self.ff(t[:, 0], *params), 0.)
        t_denom = tr_denom - tl_denom
        return t_denom

    def neg_ll(self, x, c, n, inf_c_flags, t, t_flags, gamma, p, f0, *params):
        x = copy(x) - gamma

        if 2 in c:
            like_i = self.log_like_i(x, c, n, inf_c_flags, p, f0, *params)
            x_ = copy(x[:, 0])
        else:
            like_i = 0
            x_ = copy(x)

        like = self.log_like(x_, c, n, p, f0, *params)
        like = like + like_i
        like = like - np.log(self.like_t(t, t_flags, *params))
        like = np.multiply(n, like)
        like = -np.sum(like)
        return like

    def neg_mean_D(self, x, c, n, *params):
        mask = c == 0
        x_obs = x[mask]
        n_obs = n[mask]

        gamma = 0

        # Assumes already ordered
        F = self.ff(x_obs - gamma, *params)
        D0 = F[0]
        Dn = 1 - F[-1]
        D = np.diff(F)
        D = np.concatenate([np.array([D0]), D.T, np.array([Dn])]).T

        Dr = self.sf(x[c == 1] - gamma, *params)
        Dl = self.ff(x[c == -1] - gamma, *params)

        if (n_obs > 1).any():
            n_ties = (n_obs - 1).sum()
            Df = self.df(x_obs - gamma, *params)
            # Df = Df[Df != 0]
            LL = np.concatenate([Dl, Df, Dr])
            ll_n = np.concatenate([n[c == -1], (n_obs - 1), n[c == 1]])
        else:
            Df = []
            n_ties = n_obs.sum()
            LL = np.concatenate([Dl, Dr])
            ll_n = np.concatenate([n[c == -1], n[c == 1]])

        M = np.log(D)
        M = -np.sum(M)/(M.shape[0])

        LL = -(np.log(LL) * ll_n).sum()/(n.sum() - n_obs.sum() + n_ties)
        return M + LL

    def _moment(self, n, *params, offset=False):
        if offset:
            gamma = params[0]
            params = params[1::]

            def fun(x):
                return x**n * self.df((x - gamma), *params)

            m = quad(fun, gamma, np.inf)[0]
        else:
            if hasattr(self, 'moment'):
                m = self.moment(n, *params)
            else:

                def fun(x):
                    return x**n * self.df(x, *params)

                m = quad(fun, *self.support)[0]
        return m

    def mom_moment_gen(self, *params, offset=False):
        if offset:
            k = self.k + 1
        else:
            k = self.k
        moments = np.zeros(k)
        for i in range(0, k):
            n = i + 1
            moments[i] = self._moment(n, *params, offset=offset)
        return moments

    def fit(self, x=None, c=None, n=None, t=None, how='MLE',
            offset=False, zi=False, lfp=False, tl=None, tr=None,
            xl=None, xr=None, fixed=None, heuristic='Turnbull',
            init=[], rr='y', on_d_is_0=False,
            turnbull_estimator='Fleming-Harrington'):

        r"""

        The central feature to SurPyval's capability. This function aimed to
        have an API to mimic the simplicity of the scipy API. That is, to use
        a simple :code:`fit()` call, with as many or as few parameters as
        is needed.

        Parameters
        ----------

        x : array like, optional
            Array of observations of the random variables. If x is
            :code:`None`, xl and xr must be provided.
        c : array like, optional
            Array of censoring flag. -1 is left censored, 0 is observed, 1 is
            right censored, and 2 is intervally censored. If not provided
            will assume all values are observed.
        n : array like, optional
            Array of counts for each x. If data is proivded as counts, then
            this can be provided. If :code:`None` will assume each
            observation is 1.
        t : 2D-array like, optional
            2D array like of the left and right values at which the
            respective observation was truncated. If not provided it assumes
            that no truncation occurs.
        how : {'MLE', 'MPP', 'MOM', 'MSE', 'MPS'}, optional
            Method to estimate parameters, these are:

                - MLE : Maximum Likelihood Estimation
                - MPP : Method of Probability Plotting
                - MOM : Method of Moments
                - MSE : Mean Square Error
                - MPS : Maximum Product Spacing

        offset : boolean, optional
            If :code:`True` finds the shifted distribution. If not provided
            assumes not a shifted distribution. Only works with distributions
            that are supported on the half-real line.

        tl : array like or scalar, optional
            Values of left truncation for observations. If it is a scalar
            value assumes each observation is left truncated at the value.
            If an array, it is the respective 'late entry' of the observation

        tr : array like or scalar, optional
            Values of right truncation for observations. If it is a scalar
            value assumes each observation is right truncated at the value.
            If an array, it is the respective right truncation value for each
            observation

        xl : array like, optional
            Array like of the left array for 2-dimensional input of x. This
            is useful for data that is all intervally censored. Must be used
            with the :code:`xr` input.

        xr : array like, optional
            Array like of the right array for 2-dimensional input of x. This
            is useful for data that is all intervally censored. Must be used
            with the :code:`xl` input.

        fixed : dict, optional
            Dictionary of parameters and their values to fix. Fixes parameter
            by name.

        heuristic : {'"Blom", "Median", "ECDF", "Modal", "Midpoint", "Mean",
                      "Weibull", "Benard", "Beard", "Hazen", "Gringorten",
                      "None", "Tukey", "DPW", "Fleming-Harrington",
                      "Kaplan-Meier", "Nelson-Aalen", "Filliben",
                      "Larsen", "Turnbull"}
            Plotting method to use, if using the probability plotting,
            MPP, method.

        init : array like, optional
            initial guess of parameters. Useful if method is failing.

        rr : ('y', 'x')
            The dimension on which to minimise the spacing between the line
            and the observation. If 'y' the mean square error between the
            line and vertical distance to each point is minimised. If 'x' the
            mean square error between the line and horizontal distance to each
            point is minimised.

        on_d_is_0 : boolean, optional
            For the case when using MPP and the highest value is right
            censored, you can choosed to include this value into the
            regression analysis or not. That is, if :code:`False`, all values
            where there are 0 deaths are excluded from the regression. If
            :code:`True` all values regardless of whether there is a death
            or not are included in the regression.

        turnbull_estimator : ('Nelson-Aalen', 'Kaplan-Meier', or
                              'Fleming-Harrington'), str, optional
            If using the Turnbull heuristic, you can elect to use either the
            KM, NA, or FH estimator with the Turnbull estimates of r, and d.
            Defaults to FH.

        Returns
        -------

        model : Parametric
            A parametric model with the fitted parameters and methods for
            all functions of the distribution using the fitted parameters.

        Examples
        --------
        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> x = Weibull.random(100, 10, 4)
        >>> model = Weibull.fit(x)
        >>> print(model)
        Parametric SurPyval Model
        =========================
        Distribution        : Weibull
        Fitted by           : MLE
        Parameters          :
             alpha: 10.551521182640098
              beta: 3.792549834495306
        >>> Weibull.fit(x, how='MPS', fixed={'alpha' : 10})
        Parametric SurPyval Model
        =========================
        Distribution        : Weibull
        Fitted by           : MPS
        Parameters          :
             alpha: 10.0
              beta: 3.4314657446866836
        >>> Weibull.fit(xl=x-1, xr=x+1, how='MPP')
        Parametric SurPyval Model
        =========================
        Distribution        : Weibull
        Fitted by           : MPP
        Parameters          :
             alpha: 9.943092756713078
              beta: 8.613016934518258
        >>> c = np.zeros_like(x)
        >>> c[x > 13] = 1
        >>> x[x > 13] = 13
        >>> c = c[x > 6]
        >>> x = x[x > 6]
        >>> Weibull.fit(x=x, c=c, tl=6)
        Parametric SurPyval Model
        =========================
        Distribution        : Weibull
        Fitted by           : MLE
        Parameters          :
             alpha: 10.363725328793413
              beta: 4.9886821457305865
        """

        if offset and self.name in ['Normal', 'Beta', 'Uniform',
                                    'Gumbel', 'Logistic']:
            detail = '{} distribution cannot be offset'.format(self.name)
            raise ValueError(detail)

        if how not in PARA_METHODS:
            raise ValueError('"how" must be one of: ' + str(PARA_METHODS))

        if how == 'MPP' and self.name == 'ExpoWeibull':
            detail = 'ExpoWeibull distribution does not work' + \
                     ' with probability plot fitting'
            raise ValueError(detail)

        if t is not None and how == 'MPS':
            detail = 'Maximum product spacing doesn\'t yet support tuncation'
            raise NotImplementedError(detail)

        if t is not None and how == 'MSE':
            detail = 'Mean square error doesn\'t yet support tuncation'
            raise NotImplementedError(detail)

        if t is not None and how == 'MOM':
            detail = 'Maximum product spacing doesn\'t support tuncation'
            raise ValueError(detail)

        if (lfp or zi) & (how != 'MLE'):
            detail = 'Limited failure or zero-inflated models' + \
                     ' can only be made with MLE'
            raise ValueError(detail)

        if (zi & (self.support[0] != 0)):
            detail = "zero-inflated models can only work" + \
                     "with models starting at 0"
            raise ValueError()

        x, c, n, t = surpyval.xcnt_handler(x=x, c=c, n=n, t=t,
                                           tl=tl, tr=tr, xl=xl, xr=xr)

        if (c == 1).all():
            raise ValueError("Cannot have only right censored data")

        if (c == -1).all():
            raise ValueError("Cannot have only left censored data")

        if surpyval.utils.check_no_censoring(c) and (how == 'MOM'):
            raise ValueError('Method of moments doesn\'t support censoring')

        if (surpyval.utils.no_left_or_int(c)) and \
           (how == 'MPP') and \
           (not heuristic == 'Turnbull'):
            detail = 'Probability plotting estimation with left or ' + \
                     'interval censoring only works with Turnbull heuristic'
            raise ValueError()

        if (heuristic == 'Turnbull') and \
           (not ((-1 in c) or (2 in c))) and \
           ((~np.isfinite(t[:, 1])).any()):
            # The Turnbull method is extremely memory intensive.
            # So if no left or interval censoring and no right-truncation
            # then this is equivalent.
            heuristic = turnbull_estimator

        if (not offset) & (not zi):
            detail_template = """
            Some of your data is outside support of distribution, observed
            values must be within [{lower}, {upper}].

            Are some of your observed values 0, -Inf, or Inf?
            """

            if x.ndim == 2:
                if ((x[:, 0] <= self.support[0]) & (c == 0)).any():
                    detail = detail_template.format(lower=self.support[0],
                                                    upper=self.support[1])
                    raise ValueError(detail)
                elif ((x[:, 1] >= self.support[1]) & (c == 0)).any():
                    detail = detail_template.format(lower=self.support[0],
                                                    upper=self.support[1])
                    raise ValueError(detail)
            else:
                if ((x <= self.support[0]) & (c == 0)).any():
                    detail = detail_template.format(lower=self.support[0],
                                                    upper=self.support[1])
                    raise ValueError(detail)
                elif ((x >= self.support[1]) & (c == 0)).any():
                    detail = detail_template.format(lower=self.support[0],
                                                    upper=self.support[1])
                    raise ValueError(detail)

        # Passed checks
        data = {
            'x': x,
            'c': c,
            'n': n,
            't': t
        }

        model = para.Parametric(self, how, data, offset, lfp, zi)
        fitting_info = {}

        if how != 'MPP':
            transform, inv_trans, funcs, inv_f = bounds_convert(x,
                                                                model.bounds)
            const, fixed_idx, not_fixed = fix_idx_and_function(fixed,
                                                               model.param_map,
                                                               funcs)

            fitting_info['transform'] = transform
            fitting_info['inv_trans'] = inv_trans
            fitting_info['funcs'] = funcs
            fitting_info['inv_f'] = inv_f

            fitting_info['const'] = const
            fitting_info['fixed_idx'] = fixed_idx
            fitting_info['not_fixed'] = not_fixed

            # This initial estimation is just terrible
            if init == []:
                if self.name in ['Gumbel', 'Beta', 'Normal', 'Uniform']:
                    with np.errstate(all='ignore'):
                        init = np.array(self._parameter_initialiser(x, c, n))
                else:
                    with np.errstate(all='ignore'):
                        if x.ndim == 2:
                            x_arr = x[:, 0]
                            init_mask = np.logical_or(x_arr <= self.support[0],
                                                      x_arr >= self.support[1])
                            init_mask = ~np.logical_and(init_mask, c == 0)
                            xl = x[init_mask, 0]
                            xr = x[init_mask, 1]
                            x_init = np.vstack([xl, xr]).T
                        else:
                            init_mask = np.logical_or(x <= self.support[0],
                                                      x >= self.support[1])
                            init_mask = ~np.logical_and(init_mask, c == 0)

                            x_init = x[init_mask]

                        c_init = c[init_mask]
                        n_init = n[init_mask]

                        init = self._parameter_initialiser(x_init,
                                                           c_init,
                                                           n_init,
                                                           offset=offset)
                        init = np.array(init)
                        if offset:
                            init[0] = x.min() - 1.

                if lfp:
                    _, _, _, F = pp(x, c, n, heuristic='Nelson-Aalen')

                    max_F = np.max(F)

                    if max_F > 0.5:
                        init = np.concatenate([init, [0.99]])
                    else:
                        init = np.concatenate([init_from_bounds(self),
                                              [max_F]])

                if zi:
                    init = np.concatenate([init, [(n[x == 0]).sum()/n.sum()]])

            init = transform(init)
            init = init[not_fixed]
            fitting_info['init'] = init
        else:
            # Probability plotting method does not need an initial estimate
            fitting_info['rr'] = rr
            fitting_info['heuristic'] = heuristic
            fitting_info['on_d_is_0'] = on_d_is_0
            fitting_info['turnbull_estimator'] = turnbull_estimator
            fitting_info['init'] = None

        model.fitting_info = fitting_info

        results = METHOD_FUNC_DICT[how](model)

        for k, v in results.items():
            setattr(model, k, v)

        if hasattr(model, 'params'):
            for k, v in zip(self.param_names, model.params):
                setattr(model, k, v)

        return model

    def fit_from_df(self, df, x=None, c=None, n=None,
                    xl=None, xr=None, tl=None, tr=None,
                    **fit_options):
        r"""
        The central feature to SurPyval's capability. This function aimed to
        have an API to mimic the simplicity of the scipy API. That is, to use
        a simple :code:`fit()` call, with as many or as few parameters as
        is needed.

        Parameters
        ----------

        df : DataFrame
            DataFrame of data to be used to create surpyval model

        x : string, optional
            column name for the column in df containing the variable data.
            If not provided must provide both xl and xr.

        c : string, optional
            column name for the column in df containing the censor flag of x.
            If not provided assumes all values of x are observed.

        n : string, optional
            column name in for the column in df containing the counts of x.
            If not provided assumes each x is one observation.

        tl : string or scalar, optional
            If string, column name in for the column in df containing the left
            truncation data. If scalar assumes each x is left truncated by
            that value. If not provided assumes x is not left truncated.

        tr : string or scalar, optional
            If string, column name in for the column in df containing the
            right truncation data. If scalar assumes each x is right truncated
            by that value. If not provided assumes x is not right truncated.

        xl : string, optional
            column name for the column in df containing the left interval for
            interval censored data. If left interval is -Inf, assumes left
            censored. If xl[i] == xr[i] assumes observed. Cannot be provided
            with x, must be provided with xr.

        xr : string, optional
            column name for the column in df containing the right interval
            for interval censored data. If right interval is Inf, assumes
            right censored. If xl[i] == xr[i] assumes observed. Cannot be
            provided with x, must be provided with xl.

        fit_options : dict, optional
            dictionary of fit options that will be passed to the :code:`fit`
            method, see that method for options.

        Returns
        -------

        model : Parametric
            A parametric model with the fitted parameters and methods for
            all functions of the distribution using the fitted parameters.


        Examples
        --------
        >>> import surpyval as surv
        >>> df = surv.datasets.BoforsSteel.df
        >>> model = surv.Weibull.fit_from_df(df, x='x', n='n', offset=True)
        >>> print(model)
        Parametric SurPyval Model
        =========================
        Distribution        : Weibull
        Fitted by           : MLE
        Offset (gamma)      : 39.76562962867477
        Parameters          :
             alpha: 7.141925216146524
              beta: 2.6204524040137844
        """

        if not type(df) == pd.DataFrame:
            raise ValueError("df must be a pandas DataFrame")

        if (x is not None) and ((xl is not None) or (xr is not None)):
            raise ValueError("Cannot use `x` and (`xl` and `xr`) together")

        if x is not None:
            x = df[x].astype(float)
        else:
            xl = df[xl].astype(float)
            xr = df[xr].astype(float)
            x = np.vstack([xl, xr]).T

        if c is not None:
            c = df[c].values.astype(int)

        if n is not None:
            n = df[n].values.astype(int)

        if tl is not None:
            if type(tl) == str:
                tl = df[tl].values.astype(float)
            elif np.isscalar(tl):
                tl = (np.ones(df.shape[0]) * tl).astype(float)
            else:
                raise ValueError('`tl` must be scalar or column label string')
        else:
            tl = np.ones(df.shape[0]) * -np.inf

        if tr is not None:
            if type(tr) == str:
                tr = df[tr].values.astype(float)
            elif np.isscalar(tr):
                tr = (np.ones(df.shape[0]) * tr).astype(float)
            else:
                detail = "`tr` must be scalar or a column label string"
                raise ValueError(detail)
        else:
            tr = np.ones(df.shape[0]) * np.inf

        t = np.vstack([tl, tr]).T

        return self.fit(x=x, c=c, n=n, t=t, **fit_options)

    def from_params(self, params, gamma=None, p=None, f0=None):
        r"""

        Creating a SurPyval Parametric class with provided parameters.

        Parameters
        ----------

        params : array like
            array of the parameters of the distribution.

        gamma : scalar, optional
            offset value for the distribution. If not provided will fit a
            regular, unshifted/not offset, distribution.


        Returns
        -------

        model : Parametric
            A parametric model with the fitted parameters and methods for all
            functions of the distribution using the
            fitted parameters.


        Examples
        --------
        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 4])
        >>> print(model)
        Parametric SurPyval Model
        =========================
        Distribution        : Weibull
        Fitted by           : given parameters
        Parameters          :
             alpha: 10
              beta: 4
        >>> model = Weibull.from_params([10, 4], gamma=2)
        >>> print(model)
        Parametric SurPyval Model
        =========================
        Distribution        : Weibull
        Fitted by           : given parameters
        Offset (gamma)      : 2
        Parameters          :
             alpha: 10
              beta: 4
        """
        if self.k != len(params):
            msg_base = "Must have {k} params for {dist} distribution"
            detail = msg_base.format(k=self.k, dist=self.name)
            raise ValueError(detail)

        if gamma is not None and self.name in ['Normal', 'Beta', 'Uniform',
                                               'Gumbel', 'Logistic']:
            msg_base = '{dist} distribution cannot be offset'
            detail = msg_base.format(dist=self.name)
            raise ValueError(detail)

        if gamma is not None:
            offset = True
        else:
            offset = False
            gamma = 0

        if p is not None:
            lfp = True
        else:
            lfp = False
            p = 1

        if f0 is not None:
            zi = True
        else:
            zi = False
            f0 = 0

        model = para.Parametric(self, 'given parameters',
                                None, offset, lfp, zi)
        model.gamma = gamma
        model.p = p
        model.f0 = f0
        model.params = np.array(params)

        for i, (low, upp) in enumerate(self.bounds):
            if low is None:
                lower_limit = -np.inf
            else:
                lower_limit = low
            if upp is None:
                upper_limit = np.inf
            else:
                upper_limit = upp

            if not ((lower_limit < params[i]) & (params[i] < upper_limit)):
                params = ', '.join(self.param_names)
                base = "Params {params} must be in bounds {bounds}"
                detail = base.format(params=params, bounds=self.bounds)
                raise ValueError(detail)
        model.dist = self
        return model



from autograd import numpy as np

class PowerExponential_():
    def __init__(self):
        self.name = 'PowerExponential'
        self.phi_param_map = {'c' : 0, 'a' : 1, 'n' : 2}
        self.phi_bounds = ((0, None), (None, None), (None, None))

    def phi(self, X, *params):
        X = np.atleast_2d(X)
        X1 = X[:, 0]
        X2 = X[:, 1]
        c = params[0]
        a = params[1]
        n = params[2]
        return c * np.exp(a/X1) * X2**n

    def phi_init(self, life, X):
        A = np.atleast_2d(X)
        A = np.hstack([np.ones(X.shape[0]).reshape(-1, 1), X])
        A[:, 1] = 1./A[:, 1]
        A[:, 2] = np.log(A[:, 2])
        y = np.log(life)
        c, a, n = np.linalg.lstsq(A, y, rcond=None)[0]
        return [np.exp(c), a, n]

PowerExponential = PowerExponential_()


from surpyval import np
from scipy.stats import uniform
from scipy.special import ndtri as z

from autograd import hessian
from autograd.numpy.linalg import pinv
from scipy.optimize import minimize

from surpyval import parametric as para
from surpyval.parametric.parametric_fitter import ParametricFitter

class Uniform_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        # Set 'k', the number of parameters
        self.k = 2
        self.bounds = ((None, None), (None, None),)
        self.support = (-np.inf, np.inf)
        self.plot_x_scale = 'linear'
        self.y_ticks = np.linspace(0, 1, 21)[1:-1]
        self.param_names = ['a', 'b']
        self.param_map = {
            'a' : 0,
            'b' : 1
        }

    def _parameter_initialiser(self, x, c=None, n=None):
        return np.min(x) - 1., np.max(x) + 1.

    def sf(self, x, a, b):
        r"""

        Surival (or Reliability) function for the Uniform Distribution:

        .. math::
            R(x) = \frac{b - x}{b - a}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Uniform.sf(x, 0, 6)
        array([0.83333333, 0.66666667, 0.5       , 0.33333333, 0.16666667])
        """
        return 1 - self.ff(x, a, b)

    def cs(self, x, X, a, b):
        r"""

        Surival (or Reliability) function for the Uniform Distribution:

        .. math::
            R(x, X) = \frac{R(x + X)}{R(X)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        cs : scalar or numpy array 
            The value(s) of the conditional survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Uniform.cs(x, 4, 0, 10)
        array([0.83333333, 0.66666667, 0.5       , 0.33333333, 0.16666667])
        """
        return self.sf(x + X, a, b) / self.sf(X, a, b)

    def ff(self, x, a, b):
        r"""

        Failure (CDF or unreliability) function for the Uniform Distribution:

        .. math::
            F(x) = \frac{x - a}{b - a}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        ff : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Uniform.sf(x, 0, 6)
        array([0.16666667, 0.33333333, 0.5       , 0.66666667, 0.83333333])
        """
        f = np.zeros_like(x)
        f = np.where(x < a, 0, f)
        f = np.where(x > b, 1, f)
        f = np.where(((x <= b) & (x >= a)), (x - a)/(b - a), f)
        return f

    def df(self, x, a, b):
        r"""

        Failure (CDF or unreliability) function for the Uniform Distribution:

        .. math::
            f(x) = \frac{1}{b - a}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the density function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Uniform.df(x, 0, 6)
        array([0.16666667, 0.16666667, 0.16666667, 0.16666667, 0.16666667])
        """
        d = np.zeros_like(x)
        d = np.where(x < a, 0, d)
        d = np.where(x > b, 0, d)
        d = np.where(((x <= b) & (x >= a)), 1./(b - a), d)
        return d

    def hf(self, x, a, b):
        r"""

        Instantaneous hazard rate for the Uniform Distribution:

        .. math::
            h(x) = \frac{1}{b - x}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Uniform.hf(x, 0, 6)
        array([0.2       , 0.25      , 0.33333333, 0.5       , 1.        ])
        """
        return self.df(x, a, b) / self.sf(x, a, b)

    def Hf(self, x, a, b):
        r"""

        Instantaneous hazard rate for the Uniform Distribution:

        .. math::
            H(x) = \ln \left ( b - a \right ) - \ln \left ( b - x \right )

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        hf : scalar or numpy array 
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Uniform.Hf(x, 0, 6)
        array([0.18232156, 0.40546511, 0.69314718, 1.09861229, 1.79175947])
        """
        return -np.log(self.sf(x, a, b))

    def qf(self, p, a, b):
        r"""

        Quantile function for the Uniform Distribution:

        .. math::
            q(p) = a + p(b - a)

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated 
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the Uniform distribution at each value p.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Uniform.qf(p, 0, 6)
        array([0.6, 1.2, 1.8, 2.4, 3. ])
        """
        return a + p*(b - a)

    def mean(self, a, b):
        r"""

        Mean of the Uniform distribution

        .. math::
            E = \frac{1}{2} \left ( a + b \right )

        Parameters
        ----------

        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        mean : scalar or numpy array 
            The mean(s) of the Uniform distribution

        Examples
        --------
        >>> from surpyval import Uniform
        >>> Uniform.mean(0, 6)
        3.0
        """
        return 0.5 * (a + b)

    def moment(self, n, a, b):
        r"""

        n-th (non central) moment of the Uniform distribution

        .. math::
            M(n) = \frac{1}{n +1} \sum_{i=0}^{n}a^ib^{n-i}

        Parameters
        ----------

        n : integer or numpy array of integers
            The ordinal of the moment to calculate
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        moment : scalar or numpy array 
            The moment(s) of the Uniform distribution

        Examples
        --------
        >>> from surpyval import Uniform
        >>> Uniform.moment(2, 0, 6)
        12.0
        """
        if n == 0:
            return 1
        else:
            out = np.zeros(n)
            for i in range(n):
                out[i] = a**i * b**(n-i)
            return np.sum(out)/(n + 1)

    def p(self, c, n):
        return 1 - 2 * (1 + c)**(1. - n) + (1 + 2*c)**(1. - n)

    def random(self, size, a, b):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        a : numpy array or scalar
            The lower parameter for the Uniform distribution
        b : numpy array or scalar
            The upper parameter for the Uniform distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Uniform
        >>> Uniform.random(10, 0, 6)
        array([3.50214341, 3.7978912 , 5.12238656, 4.27185221, 3.05507685,
               2.71236199, 4.89311322, 1.11373047, 4.90549424, 1.76321338])
        >>> Uniform.random((5, 5), 0, 6)
        array([[4.76809829, 4.42155933, 2.59469997, 4.31525748, 5.53469545],
               [0.06222942, 1.26267164, 1.74188626, 1.05235807, 0.92461476],
               [2.06215303, 0.02184135, 0.97058002, 3.02219656, 3.22137982],
               [2.14951891, 3.18096661, 2.37105309, 0.65710124, 0.68828779],
               [0.58827207, 3.7633596 , 5.62330526, 5.24481753, 4.23162212]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, a, b)

    def ab_cb(self, x, a, b, N, alpha=0.05):
        # Parameter confidence intervals from here:
        # https://mathoverflow.net/questions/278675/confidence-intervals-for-the-endpoints-of-the-uniform-distribution
        #
        sample_range = np.max(x) - np.min(x)
        fun = lambda c : self.p(c, N)
        c_hat = minimize(fun, 1.).x
        return a - c_hat*sample_range, b + c_hat*sample_range

    def mle(self, x, c, n, t, const, trans, inv_fs, init, fixed_idx, offset):
        params = np.array([np.min(x), np.max(x)])
        results = {}
        results['params'] = params
        return results

    def mpp_x_transform(self, x):
        return x

    def mpp_y_transform(self, y, *params):
        return y

    def mpp_inv_y_transform(self, y, *params):
        return y

    def unpack_rr(self, params, rr):
        if rr == 'y':
            a = -params[1]/params[0]
            b = (1 - params[1])/params[0]
        if rr == 'x':
            a = params[1]
            b = params[0] + params[1]

        return a, b

    def _mom(self, x):
        mu_1 = np.mean(x)
        mu_2 = np.mean(x**2)

        d = np.sqrt(3 * (mu_2 - mu_1**2))
        a = mu_1 - d
        b = mu_1 + d
        return a, b

Uniform = Uniform_('Uniform')


from autograd import numpy as np

class InversePower_():
    def __init__(self):
        self.name = 'InversePower'
        self.phi_param_map = {'a' : 0, 'n' : 1}
        self.phi_bounds = ((0, None), (None, None),)

    def phi(self, X, *params):
        return 1./(params[0] * X**params[1])

    def phi_init(self, life, X):
        X = X.flatten()
        n, a = (np.polyfit(np.log(X), np.log(1./life), 1))
        return [np.exp(a), n]

InversePower = InversePower_()

class Power_():
    def __init__(self):
        self.name = 'Power'
        self.phi_param_map = {'a' : 0, 'n' : 1}
        self.phi_bounds = ((0, None), (None, None),)

    def phi(self, X, *params):
        return params[0] * X**params[1]

    def phi_init(self, life, X):
        X = X.flatten()
        n, a = (np.polyfit(np.log(X), np.log(life), 1))
        return [np.exp(a), n]

Power = Power_()


"""
This code was created for and sponsored by Cartiga (www.cartiga.com). 
Cartiga makes no representations or warranties in connection with the code 
and waives any and all liability in connection therewith. Your use of the 
code constitutes acceptance of these terms.

Copyright 2022 Cartiga LLC
"""

from copy import copy
import numpy as np
from numpy.linalg import inv
from scipy.stats import norm

from ..parametric.fitters import bounds_convert, fix_idx_and_function
from ..nonparametric import (
    NelsonAalen,
    KaplanMeier, 
    FlemingHarrington,
    Turnbull
)

nonparametric_dists = {
    'Nelson-Aalen' : NelsonAalen,
    'Kaplan-Meier' : KaplanMeier,
    'Fleming-Harrington' : FlemingHarrington,
    'Turnbull' : Turnbull
}

from numba import njit
import numpy_indexed as npi
from scipy.optimize import root
import numpy.ma as ma

from .proportional_hazards import ProportionalHazardsModel
from .regression import Regression
from ..utils import (
    validate_coxph,
    validate_coxph_df_inputs
)

@njit
def efron_jit(n_d, Ri, Di, out):
    for i in range(len(n_d)):
        val = 0.
        if n_d[i] == 0:
            continue
        for j in range(int(n_d[i])):
            c = j / n_d[i]
            v = Ri[i] - (c * Di[i])
            val += np.log(v)[0]
        out[i] = val
    return out

# @njit
# def efron_jac_jit(n_d, Ri, ZRi, Di, ZDi, out):
#     for i in range(len(n_d)):
#         val = np.zeros(out.shape[1])
    
#         if n_d[i] == 0:
#             continue
#         for j in range(int(n_d[i])):
#             c = j / n_d[i]
#             val = val + ((ZRi[i] - (c * ZDi[i])) / (Ri[i] - (c * Di[i])))
#         out[i] = val
#     return out

def efron_jac(n_d, Ri, ZRi, Di, ZDi, masked_array):
    """
    Vectorised implementation of term two of the efron ll
    jacobian.

    This implementation runs the risk of large memory usage
    given the 3D array that is created.
    """
    r = masked_array / n_d
    
    denom = Ri - Di * r
    denom = np.expand_dims(denom, axis=-1)

    r = np.expand_dims(r, axis=-1)
    numer = np.expand_dims(ZDi, axis=1) * r
    numer = (np.expand_dims(ZRi, axis=1) - numer)
    
    out = numer / denom
    out = out.sum(axis=1)
    return out

@njit
def efron_hess_jit(n_d, Ri, ZRi, Z2Ri, Di, ZDi, Z2Di, out):
    # TODO: Vectorised implementation like for the jacobian above
    # Not sure my brain can handle that many dimensions.
    for i in range(len(n_d)):
        val = np.zeros(out.shape[1:])
        if n_d[i] == 0:
            continue
        for j in range(int(n_d[i])):
            c = j / n_d[i]
            dRD = (Ri[i] - c * Di[i])
            a = (ZRi[i] - c * ZDi[i])
            # If einsum ever supported by numba, remove this
            if a.shape == (1,):
                a2 = a * a
            else:
                a2 = a @ a.reshape(-1, 1)
            val += ( 
                   + dRD * (Z2Ri[i] - c * Z2Di[i])
                   # Einsum not supported by numba
                   # Change if this ever is the case.
                   # - np.einsum('i, j -> ij', a, a)
                   - a2
                   ) / dRD**2
        out[i] = val
    return out

def at_risk_beta_Z(arr, n, gb_x):    
    R = gb_x.sum(n * arr)[1]
    # Get the reverse cumulative sum
    return R[::-1].cumsum(axis=0)[::-1]

class CoxPH_():
    """
    Best reference I can find that covers all the 
    possibilities for estimating betas
    http://www-personal.umich.edu/~yili/lect4notes.pdf
    """

    def baseline(self, beta, x, c, n, Z):
        """
        Uses the Breslow method to compute the baseline hazard.
        """
        unique_x = np.unique(x)

        d = np.zeros_like(unique_x)
        r = np.zeros_like(unique_x)
        
        for i, tau_i in enumerate(unique_x):
            mask_d_i = (x == tau_i) & (c == 0)
            d[i] = n[mask_d_i].sum()

            mask_at_risk_i = (x >= tau_i)
            Z_ri = Z[mask_at_risk_i, :]

            r[i] = np.exp(Z_ri @ beta).sum()

        return unique_x, r, d



    def create_efron_ll_jac_hess(self, x, Z, c, n, tl,
                                   with_jac=True,
                                   with_hess=True):
        """
        The reference used to compute the jacobian and hessian
        was https://mathweb.ucsd.edu/~rxu/math284/slect5.pdf
        """
        # TODO: Incorporate left-truncation... somehow.
        # left truncation allows for implementation of time-varying covariates

        # To do left truncation. All one needs to do is to adjust the risk set.
        # I am aiming to implement this by:
        # Ri - TRi.. that's it!

        # Groupby object for repeated use
        gb_x = npi.group_by(x)
        gb_tl = npi.group_by(tl)
        n_d_x = np.where(c == 0, n, 0)
        n_d = gb_x.sum(n_d_x)[1]
        n_d_x = n_d_x.reshape(-1, 1)
        n = n.reshape(-1, 1)

        max_n = n_d.max()

        x_ = gb_x.unique
        x_tl = gb_tl.unique
        idx = np.searchsorted(x_, x_tl, side='right')
        
        def log_like(beta):
            beta_z = Z @ beta
            
            S_d = gb_x.sum(n_d_x * beta_z.reshape(-1, 1))[1].reshape(-1, 1)
            e_beta_z = np.exp(beta_z).reshape(-1, 1)
            
            x_, Ri = gb_x.sum(n * e_beta_z)
            trunc = gb_tl.sum(n * e_beta_z)[1]
            
            Ri = Ri[::-1].cumsum(axis=0)[::-1]
            
            trunc = (trunc[::-1].cumsum(axis=0)[::-1] - trunc)

            TRi = np.ones_like(Ri) * np.inf
            
            TRi[idx] = trunc
            TRi = np.minimum.accumulate(TRi)
            # Subtract the truncated values from the others.
            Ri = Ri - TRi

            Di = gb_x.sum(n_d_x * e_beta_z)[1]
            
            efron_denom = np.zeros_like(x_)
            efron_denom = efron_jit(n_d, Ri, Di, efron_denom)
            
            like = S_d.sum() - efron_denom.sum()
            return -like

        S_d = gb_x.sum(n_d_x * Z)[1]

        arr = np.repeat([np.arange(max_n)], len(n_d), axis=0)
        mask = 1 - (arr < n_d.reshape(-1, 1)).astype(int)

        masked_array = ma.array(arr, mask=mask)
        
        def jac_hess(beta):
            # This line troubled me for longer than I care
            # to admit. I was using n, but it is only the
            # number of deaths at each point, n_d_x

            # Z = Z
            Z2 = np.einsum('ij, ik -> ijk', Z, Z)

            # Only call this once.. Yay.
            beta_z = Z @ beta

            e_beta_z = np.exp(beta_z).reshape(-1, 1)
            z_e_beta_z = Z * e_beta_z
            z2_e_beta_z = np.einsum('ijk, ij -> ijk', Z2, n * e_beta_z)

            Ri = at_risk_beta_Z(e_beta_z, n, gb_x)
            ZRi = at_risk_beta_Z(z_e_beta_z, n, gb_x)
            Z2Ri = gb_x.sum(z2_e_beta_z)[1]
            Z2Ri = Z2Ri[::-1].cumsum(axis=0)[::-1]

            trunc = gb_tl.sum(n * e_beta_z)[1]
            trunc = trunc[::-1].cumsum(axis=0)[::-1] - trunc

            z_trunc = gb_tl.sum(n * z_e_beta_z)[1]
            z_trunc = z_trunc[::-1].cumsum(axis=0)[::-1] - z_trunc

            z2_trunc = gb_tl.sum(z2_e_beta_z)[1]
            z2_trunc = z2_trunc[::-1].cumsum(axis=0)[::-1] - z2_trunc

            TRi = np.ones_like(Ri) * np.inf
            ZTRi = np.ones_like(ZRi) * np.inf
            Z2TRi = np.ones_like(Z2Ri) * np.inf
            
            TRi[idx] = trunc
            TRi = np.minimum.accumulate(TRi)

            ZTRi[idx] = z_trunc
            ZTRi = np.minimum.accumulate(ZTRi)

            Z2TRi[idx] = z2_trunc
            Z2TRi = np.minimum.accumulate(Z2TRi)
            
            # Subtract the truncated values from the non-truncated risk set.
            Ri -= TRi
            ZRi -= ZTRi
            Z2Ri -= Z2TRi 

            Di = gb_x.sum(n_d_x * e_beta_z)[1]
            ZDi = gb_x.sum(n_d_x * z_e_beta_z)[1]
            
            expected_S_d = np.zeros_like(S_d)
            expected_S_d = efron_jac(n_d.reshape(-1, 1), Ri, ZRi, Di, ZDi, masked_array)
            
            diff = S_d - expected_S_d
            jacobian = -diff.sum(axis=0)

            # Something remains incorrect with the Hessian
            Z2Di = np.einsum('ijk, ij -> ijk', Z2, n_d_x * e_beta_z)
            Z2Di = gb_x.sum(Z2Di)[1]

            hess_matrix = np.zeros_like(Z2)
            hess_matrix = efron_hess_jit(n_d, Ri, ZRi, Z2Ri, Di, ZDi, Z2Di, hess_matrix)
            hess_matrix = -hess_matrix.sum(axis=0)
            
            return jacobian#, hess_matrix

        return log_like, jac_hess, False

    def create_breslow_ll_jac_hess(self, x, Z, c, n, tl):
        """
        The reference used to compute the jacobian and hessian
        was https://mathweb.ucsd.edu/~rxu/math284/slect5.pdf
        """
        # TODO: Incorporate left-truncation... somehow.
        # left truncation allows for implementation of time-varying covariates

        gb_x = npi.group_by(x)
        gb_tl = npi.group_by(tl)
        n_d_x = np.where(c == 0, n, 0)
        n_d = gb_x.sum(n_d_x)[1]
        n_d_x = n_d_x.reshape(-1, 1)
        n = n.reshape(-1, 1)

        x_ = gb_x.unique
        x_tl = gb_tl.unique
        idx = np.searchsorted(x_, x_tl, side='right')
        
        # Create the log_like function for the data
        def log_like(beta):
            beta_z = Z @ beta
            di_beta_z = gb_x.sum(n_d_x * beta_z.reshape(-1, 1))[1].reshape(-1, 1)
            e_beta_z = np.exp(beta_z).reshape(-1, 1)
            Ri = at_risk_beta_Z(e_beta_z, n, gb_x)

            trunc = gb_tl.sum(n * e_beta_z)[1]
            trunc = trunc[::-1].cumsum(axis=0)[::-1] - trunc
            
            TRi = np.ones_like(Ri) * np.inf

            TRi[idx] = trunc
            TRi = np.minimum.accumulate(TRi)

            Ri -= TRi

            Ri = np.log(Ri)
            Ri = n_d.reshape(-1, 1) * Ri

            
            like = di_beta_z - Ri
            
            return -like.sum()

        S_d = gb_x.sum(n_d_x.reshape(-1, 1) * Z)[1]

        def jac_hess(beta):
            Z2 = np.einsum('ij, ik -> ijk', Z, Z)

            # Only call this once.. Yay.
            beta_z = Z @ beta

            e_beta_z = np.exp(beta_z).reshape(-1, 1)
            z_e_beta_z = Z * e_beta_z
            z2_e_beta_z = np.einsum('ijk, ij -> ijk', Z2, n * e_beta_z)

            Ri = at_risk_beta_Z(e_beta_z, n, gb_x)
            ZRi = at_risk_beta_Z(z_e_beta_z, n, gb_x)
            Z2Ri = gb_x.sum(z2_e_beta_z)[1]
            Z2Ri = Z2Ri[::-1].cumsum(axis=0)[::-1]

            trunc = gb_tl.sum(n * e_beta_z)[1]
            trunc = trunc[::-1].cumsum(axis=0)[::-1] - trunc

            z_trunc = gb_tl.sum(n * z_e_beta_z)[1]
            z_trunc = z_trunc[::-1].cumsum(axis=0)[::-1] - z_trunc

            z2_trunc = gb_tl.sum(z2_e_beta_z)[1]
            z2_trunc = z2_trunc[::-1].cumsum(axis=0)[::-1] - z2_trunc

            TRi = np.ones_like(Ri) * np.inf
            ZTRi = np.ones_like(ZRi) * np.inf
            Z2TRi = np.ones_like(Z2Ri) * np.inf
            
            TRi[idx] = trunc
            TRi = np.minimum.accumulate(TRi)

            ZTRi[idx] = z_trunc
            ZTRi = np.minimum.accumulate(ZTRi)

            Z2TRi[idx] = z2_trunc
            Z2TRi = np.minimum.accumulate(Z2TRi)
            
            # Subtract the truncated values from the non-truncated risk set.
            Ri -= TRi
            ZRi -= ZTRi
            Z2Ri -= Z2TRi

            EZ = ZRi / Ri
            EZ = n_d.reshape(-1, 1) * EZ

            jacobian = -(S_d - EZ).sum(axis=0)
            
            # calc term 1
            term_1 = np.einsum('ijk,ij->ijk', Z2Ri, 1./Ri)
            
            # calc term 2
            term_2 = ZRi / Ri
            term_2 = np.einsum('ij, ik-> ijk', term_2, term_2) 
            
            # Compute the Hessian matrix
            hess_matrix = term_1 - term_2
            hess_matrix = np.einsum('ijk,i->ijk', hess_matrix, n_d.flatten())
            hess_matrix = hess_matrix.sum(axis=0)

            return jacobian, hess_matrix            

        return log_like, jac_hess, True

    def fit(self, x, Z, c=None, n=None, tl=None, method="breslow", tol=1e-10):
        """
        Need to add shape checking of x, c, n
        """
        x, c, n, tl, Z = validate_coxph(x, c, n, Z, tl, method)

        # Good initial guess assumes no impact
        beta_init = np.zeros(Z.shape[1])

        if method == "efron":
            func_generator = self.create_efron_ll_jac_hess
        elif method == "breslow":
            func_generator = self.create_breslow_ll_jac_hess
        # TODO: Cox-Exact
        # TODO: K&P
        # See: https://mathweb.ucsd.edu/~rxu/math284/slect5.pdf

        neg_ll, jac, hess = func_generator(x, Z, c, n, tl)

        # Have found that root finding is faster than minimization
        res = root(jac, beta_init, jac=hess, tol=tol)


        # The hessian is at [1] of the jac function
        if hess:
            # Finds the p-value for the null hypothesis 
            # that the coefficient is 0.
            var = np.diag(inv(jac(res.x)[1]))
            z_score = res.x / np.sqrt(var)
            p_values = 2*(1 - norm.cdf(np.abs(z_score)))
        else:
            p_values = None


        model = Regression()
        model = ProportionalHazardsModel("Cox", "Semi-Parametric")
        model._neg_log_like = neg_ll(res.x)
        model.p_values = p_values
        model.neg_ll = neg_ll
        model.jac = jac
        model.hess = hess
        model.tie_method = method
        model.baseline_method = "breslow"
        model.res = res
        model.beta = copy(res.x)
        model.phi = lambda Z: np.exp(Z @ model.beta)
        model.parameters = res.x

        x, r, d = self.baseline(model.beta, x, c, n, Z)
        model.x = x
        model.r = r
        model.d = d
        model.tl = tl
        model.h0 = d / r
        model.H0 = model.h0.cumsum()

        return model

    def fit_from_df(self, df, x_col, Z_cols=None, c_col=None,
                    n_col=None, formula=None, method="efron"):

        x, c, n, Z, form = validate_coxph_df_inputs(
            df, 
            x_col, c_col, n_col, 
            Z_cols,
            formula
        )

        model = self.fit(x, Z, c, n, method=method)
        model.formula = form

        return model

CoxPH = CoxPH_()


import surpyval
from surpyval import np
from scipy.linalg import inv
from scipy.optimize import minimize
from scipy.special import ndtri as z

from autograd import jacobian, hessian
from surpyval import parametric as para


class MixtureModel():
    """
    Generalised from algorithm found here
    https://www.sciencedirect.com/science/article/pii/S0307904X12002545
    """
    def __init__(self, x, dist=para.Weibull, **kwargs):
        self.m = kwargs.pop('m', 2)
        raw_data = {}
        c = kwargs.pop('c', None)
        n = kwargs.pop('n', None)
        raw_data['x'] = x
        raw_data['c'] = c
        raw_data['n'] = n
        self.raw_data = raw_data
        self.dist = dist

        x, c, n = surpyval.xcn_handler(x, c, n)
        assert len(x) > self.m * (self.dist.k + 1)

        self.x = x
        self.c = c
        self.n = n

        self.N = n.sum()
        splits_x = np.array_split(x, self.m)
        splits_c = np.array_split(c, self.m)
        splits_n = np.array_split(n, self.m)
        params = np.zeros(shape=(self.m, self.dist.k))

        for i in range(self.m):
            params[i, :] = self.dist.fit(x=splits_x[i], c=splits_c[i],
                                         n=splits_n[i]).params
        self.params = params
        self.w = np.ones(shape=(self.m)) / self.m
        self.p = np.ones(shape=(self.m, len(self.x))) / self.m

    def Q(self, params):
        params = params.reshape(self.m, self.dist.k)
        f = np.zeros_like(self.p)
        for i in range(self.m):
            like = self.dist.like(self.x, self.c, self.n, *params[i])
            like += surpyval.TINIEST
            like = np.where(like < 1, like, 1)
            like = np.log(like)
            like = np.multiply(self.n, like)

            f[i] = np.multiply(self.p[i], like)
        f = -np.sum(f)
        self.loglike = f
        return f

    def expectation(self):
        for i in range(self.m):
            like = self.dist.like(self.x, self.c, self.n, *self.params[i])
            like = np.multiply(self.w[i], like)
            self.p[i] = like
        self.p = np.divide(self.p, np.sum(self.p, axis=0))
        self.w = np.sum(self.p, axis=1) / self.N

    def maximisation(self):
        bounds = self.dist.bounds * self.m
        res = minimize(self.Q, self.params, bounds=bounds)
        # if not res.success:
        #   print(res)
        #   raise Exception('Max failed')
        self.params = res.x.reshape(self.m, self.dist.k)

    def EM(self):
        self.expectation()
        self.maximisation()

    def _em(self, tol=1e-6, max_iter=1000):
        i = 0
        self.EM()
        f0 = self.loglike
        self.EM()
        f1 = self.loglike
        while (np.abs(f0 - f1) > tol) and (i < max_iter):
            f0 = f1
            self.EM()
            f1 = self.loglike
            i += 1
        if i >= 1000:
            print('Max iterations reached')

    def neg_ll(self, x, c, n, *params):
        f = np.zeros_like(self.p)
        params = np.reshape(params, (self.m, self.dist.k + 1))
        f = np.zeros_like(x)
        for i in range(self.m):
            like = self.dist.like(x, c, n, *params[i, 1::])
            like = np.multiply(params[i, 0], like)
            f = f + like
        f = np.where(f <= 0, surpyval.TINIEST, f)
        f = np.where(f < 1, f, 1)
        f = np.log(f)
        f = np.multiply(n, f)
        f = -np.sum(f)
        return f

    def _mle(self):

        def fun(x):
            return self.neg_ll(self.x, self.c, self.n, *x)

        bounds = tuple(((0, 1), *self.dist.bounds)) * self.m
        init = np.hstack([np.atleast_2d(self.w).T, self.params])

        def w_sums_to_1(x):
            xx = np.reshape(x, (self.m, self.dist.k + 1))
            return xx[:, 0].sum() - 1
        cons = {'type': 'eq', 'fun': w_sums_to_1}
        jac = jacobian(fun)
        hess = hessian(fun)
        res = minimize(fun, init.flatten(),
                       jac=jac, bounds=bounds,
                       constraints=[cons])
        self.hess_inv = inv(hess(res.x))
        self.res = res
        self.params = np.reshape(res.x, (self.m, self.dist.k + 1))[:, 1::]
        self.w = np.reshape(res.x, (self.m, self.dist.k + 1))[:, 0]

    def fit(self, how='EM'):
        if how == 'EM':
            self._em()
            self.method = 'EM'
        elif how == 'MLE':
            self._mle()
            self.method = 'MLE'

    def ll(self):
        return self.loglike

    def R_cb(self, t, cb=0.05):

        def ssf(params):
            params = np.reshape(params, (self.m, self.dist.k + 1))
            F = np.zeros_like(t)
            for i in range(self.m):
                F = F + params[i, 0] * self.dist.ff(t, *params[i, 1::])
            return 1 - F

        pvars = self.hess_inv[np.triu_indices(self.hess_inv.shape[0])]
        with np.errstate(all='ignore'):
            jac = jacobian(ssf)(self.res.x)

        var_u = []
        for i, j in enumerate(jac):
            j = np.atleast_2d(j).T * j
            j = j[np.triu_indices(j.shape[0])]
            var_u.append(np.sum(j * pvars))
        diff = (z(cb/2) * np.sqrt(np.array(var_u))
                * np.array([1., -1.]).reshape(2, 1))
        R_hat = self.sf(t)
        exponent = diff/(R_hat*(1 - R_hat))
        R_cb = R_hat / (R_hat + (1 - R_hat) * np.exp(exponent))
        return R_cb.T

    def mean(self):
        mean = 0
        for i in range(self.m):
            mean += self.w[i] * self.dist.mean(*self.params[i])
        return mean

    def random(self, size):
        sizes = np.random.multinomial(size, self.w)
        rvs = np.zeros(size)
        s_last = 0
        for i, s in enumerate(sizes):
            rvs[s_last:s+s_last] = self.dist.random(s, *self.params[i, :])
            s_last = s
        np.random.shuffle(rvs)
        return rvs

    def df(self, t):
        f = np.zeros_like(t)
        for i in range(self.m):
            f += self.w[i] * self.dist.df(t, *self.params[i])
        return f

    def ff(self, t):
        F = np.zeros_like(t)
        for i in range(self.m):
            F = F + self.w[i] * self.dist.ff(t, *self.params[i])
        return F

    def sf(self, t):
        return 1 - self.ff(t)

    def cs(self, t, T):
        return self.sf(t + T) / self.sf(T)



import pytest
import numpy as np
import lifelines
import surpyval

def right_censor(x, tl, frac):
    c = np.random.binomial(1, frac, x.shape)
    x_out = np.copy(x)
    for i, (trunc, value) in enumerate(zip(tl, x)):
        if c[i] == 0:
            continue
        if np.isfinite(trunc):
            x_out[i] = np.random.uniform(trunc, value, 1)
        else:
            x_out[i] = value - np.abs(value * np.random.uniform(0, 1, 1))
    return x_out, c

def left_truncate(x, dist, frac, params):
    t = np.random.binomial(1, frac, x.shape)
    # Find a lower value
    tl = np.where((t == 1) & (x > 0), x * np.random.uniform(0, 1, x.size), -np.inf)
    tl = np.where((t == 1) & (x < 0), x - np.abs(x * np.random.uniform(0, 1, x.size)), tl)
    drop_due_to_truncation = dist.ff(tl, *params)
    drop_due_to_truncation[~np.isfinite(drop_due_to_truncation)] = 0
    
    keep = np.ones_like(x)
    for i, p in enumerate(drop_due_to_truncation):
        if p == 0:
            continue
        else:
            keep[i] = np.random.binomial(1, p)

    mask = keep == 1
    tl = tl[mask]
    x = x[mask]
    return x, tl

def test_kaplan_meier_against_lifelines():
    kmf = lifelines.KaplanMeierFitter()
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        n = np.ones_like(x) * int(np.random.uniform(1, 5))
        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_est = kmf.fit(x, weights=n).predict(x_test).values
        surp_est = surpyval.KaplanMeier.fit(x, n=n).sf(x_test)
        if not np.allclose(ll_est, surp_est, 1e-15):
            raise AssertionError('Kaplan-Meier different to lifelines?!')

def test_kaplan_meier_censored_against_lifelines():
    kmf = lifelines.KaplanMeierFitter()
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        c = np.random.binomial(1, np.random.uniform(0, 1, 1), x.shape)
        x = x - np.abs(x * np.random.uniform(0, 1, x.shape))
        n = np.ones_like(x) * int(np.random.uniform(1, 5))
        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_est = kmf.fit(x, 1-c, weights=n).predict(x_test).values
        surp_est = surpyval.KaplanMeier.fit(x, c=c, n=n).sf(x_test)
        if not np.allclose(ll_est, surp_est, 1e-15):
            raise AssertionError('Kaplan-Meier different to lifelines?!')


def test_kaplan_meier_censored_and_truncated_against_lifelines():
    kmf = lifelines.KaplanMeierFitter()
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        x, tl = left_truncate(x, surpyval.Weibull, 0.1, test_params)
        x, c = right_censor(x, tl, 0.2)
        n = np.ones_like(x) * int(np.random.uniform(1, 5))
        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_est = kmf.fit(x, 1-c, entry=tl, weights=n).predict(x_test).values
        surp_est = surpyval.KaplanMeier.fit(x, c=c, n=n, tl=tl).sf(x_test)
        if not np.allclose(ll_est, surp_est, 1e-15):
            raise AssertionError('Kaplan-Meier different to lifelines?!')


def test_nelson_aalen_against_lifelines():
    naf = lifelines.NelsonAalenFitter(nelson_aalen_smoothing=False)
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        n = np.ones_like(x) * int(np.random.uniform(1, 5))
        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_est = naf.fit(x, weights=n).predict(x_test).values
        surp_est = surpyval.NelsonAalen.fit(x, n=n).Hf(x_test)
        if not np.allclose(ll_est, surp_est, 1e-15):
            raise AssertionError('Kaplan-Meier different to lifelines?!')

def test_nelson_aalen_censored_against_lifelines():
    naf = lifelines.NelsonAalenFitter(nelson_aalen_smoothing=False)
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        c = np.random.binomial(1, np.random.uniform(0, 1, 1), x.shape)
        x = x - np.abs(x * np.random.uniform(0, 1, x.shape))
        n = np.ones_like(x) * int(np.random.uniform(1, 5))
        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_est = naf.fit(x, 1-c, weights=n).predict(x_test).values
        surp_est = surpyval.NelsonAalen.fit(x, c=c, n=n).Hf(x_test)
        if not np.allclose(ll_est, surp_est, 1e-15):
            raise AssertionError('Kaplan-Meier different to lifelines?!')


def test_nelson_aalen_censored_and_truncated_against_lifelines():
    naf = lifelines.NelsonAalenFitter(nelson_aalen_smoothing=False)
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        x, tl = left_truncate(x, surpyval.Weibull, 0.1, test_params)
        x, c = right_censor(x, tl, 0.2)
        n = np.ones_like(x) * int(np.random.uniform(1, 5))
        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_est = naf.fit(x, 1-c, entry=tl, weights=n).predict(x_test).values
        surp_est = surpyval.NelsonAalen.fit(x, c=c, n=n, tl=tl).Hf(x_test)
        if not np.allclose(ll_est, surp_est, 1e-15):
            raise AssertionError('Kaplan-Meier different to lifelines?!')


def test_fleming_harrington_same_as_nelson_aalen_with_no_counts():
    naf = lifelines.NelsonAalenFitter(nelson_aalen_smoothing=False)
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)
        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_est = naf.fit(x).predict(x_test).values
        surp_est = surpyval.FlemingHarrington.fit(x).Hf(x_test)
        if not np.allclose(ll_est, surp_est, 1e-15):
            raise AssertionError('Fleming-Harrington fails different to lifelines?!')


def test_fleming_harrington_HF_less_than_or_equal_to_nelson_aalen_with_counts():
    naf = lifelines.NelsonAalenFitter(nelson_aalen_smoothing=False)
    for i in range(100):
        test_params = []
        for b in ((1, 100), (0.5, 20)):
            test_params.append(np.random.uniform(*b))
        test_params = np.array(test_params)

        x = surpyval.Weibull.random(int(np.random.uniform(2, 1000, 1)), *test_params)
        n = np.ones_like(x) * int(np.random.uniform(2, 5))

        x_test = np.random.uniform(x.min()/2, x.max()*2, 100)
        ll_na_est = naf.fit(x, weights=n).predict(x_test).values
        surp_est = surpyval.FlemingHarrington.fit(x, n=n).Hf(x_test)
        # FH cumulative hazard should be less than NA Hf
        diff = surp_est - ll_na_est
        if (diff < 0).all():
            raise AssertionError('Fleming-Harrington not all below NelsonAalen')

def test_fleming_harrington_case():
    solution_h = [
        (1./10 + 1./9 + 1./8),
        (1./7 + 1./6),
        (1./5 + 1./4 + 1./3 + 1./2),
        1.
    ]
    x = [1, 2, 3, 4]
    n = [3, 2, 4, 1]
    H = np.cumsum(solution_h)
    H_est = surpyval.FlemingHarrington.fit(x=x, n=n).H
    assert np.allclose(H, H_est, 1e-15)


def test_turnbull_with_R_1():
    """
    R's icenReg solution, from code:

    ``` r
    left <-  c(1, 8, 8, 7, 7, 17, 37, 46, 46, 45)
    right <- c(7, 8, 15, 16, 14, Inf, 44, Inf, Inf, Inf)

    df <- data.frame(cbind(left, right))

    r_model <- ic_np(cbind(left, right) ~ 0, df)

    getSCurves(r_model)
    ```
    """
    x_test = [1, 7, 8, 14, 44, np.inf]
    sf_test = [1., 0.9, 0.7, 0.5, 0.375, 0]

    left = np.array([1, 8, 8, 7, 7, 17, 37, 46, 46, 45])
    right = np.array([7, 8, 10, 16, 14, np.inf, 44, np.inf, np.inf, np.inf])

    model = surpyval.Turnbull.fit(xl=left,
                              xr=right,
                              turnbull_estimator='Kaplan-Meier')

    assert np.allclose(model.sf(x_test), sf_test, 1e-5)


def test_turnbull_with_R_2():
    """
    R's icenReg solution, from code:

    ``` r
    left <- c(1, 76, 288, 501, 579, 667, 829, 920, 1071)
    right <- c(71, 169, 344, 504, 579, 754, 829, 971, Inf)

    df <- data.frame(cbind(left, right))

    r_model <- ic_np(cbind(left, right) ~ 0, df)

    getSCurves(r_model)
    ```
    """
    x_test = [1, 71, 169, 344, 504, 579, 754, 829, 971, np.inf]
    sf_test = [1.0, 0.8888889, 0.7777778, 0.6666667, 0.5555556, 0.4444444,
              0.3333333, 0.2222222, 0.1111111, 0.0]


    left = [1, 76, 288, 501, 579, 667, 829, 920, 1071]
    right = [71, 169, 344, 504, 579, 754, 829, 971, np.inf]

    model = surpyval.Turnbull.fit(xl=left,
                                  xr=right,
                                  turnbull_estimator='Kaplan-Meier')

    assert np.allclose(model.sf(x_test), sf_test, 1e-5)

[0.875, 0.75, 0.625, 0.5, 0.375, 0.25, 0.125, 0.0]

def test_turnbull_with_R_3():
    """
    R's icenReg solution, from code:

    ``` r
    left <- seq(1, 10)
    right <- seq(2, 11)

    df <- data.frame(cbind(left, right))

    t <- ic_np(cbind(left, right) ~ 0, df)

    getSCurves(t)
    ```
    """
    x_test = np.array(range(1, 12))
    sf_test = [1., 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.]
    left = np.array(range(1, 11))
    right = left + 1


    model = surpyval.Turnbull.fit(xl=left,
                                  xr=right,
                                  turnbull_estimator='Kaplan-Meier')

    assert np.allclose(model.sf(x_test), sf_test, 1e-5)




def test_kaplan_meier_with_R():
    """
    R's survreg() solutions.
    """
    # Test 1
    x1 = np.array(range(1, 9))
    sf_test1 = [0.875, 0.75, 0.625, 0.5, 0.375, 0.25, 0.125, 0.0]
    model = surpyval.KaplanMeier.fit(x1)
    assert np.allclose(model.R, sf_test1, 1e-5)

    # If repeated should be same
    n = np.ones_like(x1) * 10
    model = surpyval.KaplanMeier.fit(x1, n=n)
    assert np.allclose(model.R, sf_test1, 1e-5)

    """
    times <- c(1, 2, 3, 4, 5, 6, 7, 8)
    weights <- c(1, 1, 1, 1, 1, 1, 1, 10)
    censored <- c(1, 1, 1, 1, 1, 1, 1, 0)
    model <- survfit(Surv(times, censored) ~ 0, weights=weights)
    model$surv
    [1] 0.9411765 0.8823529 0.8235294 0.7647059 0.7058824 0.6470588 0.5882353
    [8] 0.5882353
    """
    # If repeated should be same
    n = np.ones_like(x1)
    n[-1] = 10
    c = np.zeros_like(x1)
    c[-1] =1
    sf_test3 = [0.94117647, 0.88235294, 0.82352941, 0.76470588, 0.70588235,
                0.64705882, 0.58823529, 0.58823529]
    model = surpyval.KaplanMeier.fit(x1, c=c, n=n)
    assert np.allclose(model.R, sf_test3, 1e-5)


def test_nelson_aalen_with_R():
    """
    R's survreg() solutions.
    """
    """
    times <- c(1, 2, 3, 4, 5, 6, 7, 8)
    weights <- c(10, 10, 10, 10, 10, 10, 10, 10)
    model <- survfit(Surv(times) ~ 0, type='fh')
    model$surv
    [1] 0.88249690 0.76501706 0.64757296 0.53018790 0.41291075 0.29586348
    [7] 0.17945027 0.06601607
    """
    # Test 1
    x1 = np.array(range(1, 9))
    sf_test1 = [0.88249690, 0.76501706, 0.64757296, 0.53018790, 0.41291075,
                0.29586348, 0.17945027, 0.06601607]
    model = surpyval.NelsonAalen.fit(x1)
    assert np.allclose(model.R, sf_test1, 1e-5)

    # If repeated should be same
    n = np.ones_like(x1) * 10
    model = surpyval.NelsonAalen.fit(x1, n=n)
    assert np.allclose(model.R, sf_test1, 1e-5)

    """
    times <- c(1, 2, 3, 4, 5, 6, 7, 8)
    weights <- c(1, 1, 1, 1, 1, 1, 1, 10)
    censored <- c(1, 1, 1, 1, 1, 1, 1, 0)
    model <- survfit(Surv(times, censored) ~ 0, weights=weights, type='fh')
    model$surv
    [1] 0.9428731 0.8857473 0.8286228 0.7714999 0.7143789 0.6572603 0.6001448
    [8] 0.6001448
    """

    # If repeated should be same
    n = np.ones_like(x1)
    n[-1] = 10
    c = np.zeros_like(x1)
    c[-1] =1
    sf_test3 = [0.9428731, 0.8857473, 0.8286228, 0.7714999, 
                0.7143789, 0.6572603, 0.6001448, 0.6001448]
    model = surpyval.NelsonAalen.fit(x1, c=c, n=n)
    assert np.allclose(model.R, sf_test3, 1e-5)



from autograd import numpy as np

class GeneralLogLinear_():
    def __init__(self):
        self.name = 'GeneralLogLinear'
        self.phi_bounds = lambda X: (((None, None),) * X.shape[1])
        self.phi_param_map = lambda X: {'beta_' + str(i) : i for i in range(X.shape[1])}

    def phi(self, X, *params):
        return np.exp(np.dot(X, np.array(params)))

    def phi_init(self, X):
        # return np.zeros(X.shape[1])
        return 1./X.mean(axis=0)

GeneralLogLinear = GeneralLogLinear_()


from surpyval.parametric.parametric_fitter import ParametricFitter
from surpyval import np
from autograd import elementwise_grad
import inspect


class Distribution(ParametricFitter):
    def __init__(self, name, fun, param_names, bounds, support):
        if str(inspect.signature(fun)) != '(x, *params)':
            detail = 'Function must have the signature \'(x, *params)\''
            raise ValueError(detail)

        if len(param_names) != len(bounds):
            raise ValueError('param_names and bounds must have same length')

        if 'p' in param_names:
            detail = "'p' reserved parameter name for LFP distributions"
            raise ValueError(detail)

        if 'gamma' in param_names:
            detail = "'gamma' reserved parameter name for offset distributions"
            raise ValueError(detail)

        if 'f0' in param_names:
            detail = "'f0' reserved parameter name for zero" + \
                     "inflated or hurdle models"
            raise ValueError(detail)

        for p_name in param_names:
            if hasattr(self, p_name):
                detail = "Can't name a parameter after a function"
                raise ValueError(detail)

        self.name = name
        self.k = len(param_names)
        self.bounds = bounds
        self.support = support
        self.plot_x_scale = 'linear'
        self.y_ticks = np.linspace(0, 1, 11)
        self.param_names = param_names
        self.param_map = {v: i for i, v in enumerate(param_names)}
        self.Hf = fun
        self.hf = lambda x, *params: elementwise_grad(self.Hf)(x, *params)
        self.sf = lambda x, *params: np.exp(-self.Hf(x, *params))
        self.ff = lambda x, *params: 1 - np.exp(-self.Hf(x, *params))
        self.df = lambda x, *params: elementwise_grad(self.ff)(x, *params)

    def _parameter_initialiser(self, x, c=None, n=None, t=None, offset=False):
        out = []
        for low, high in self.bounds:
            if (low is None) & (high is None):
                out.append(0)
            elif high is None:
                out.append(low + 1.)
            elif low is None:
                out.append(high - 1.)
            else:
                out.append((high + low)/2.)

        return out

    def mpp_inv_y_transform(self, y, *params):
        return y

    def mpp_y_transform(self, y, *params):
        return y

    def mpp_x_transform(self, x, gamma=0):
        return x - gamma



import re
from surpyval import np
import matplotlib.pyplot as plt
from autograd import jacobian
from scipy.stats import uniform
from surpyval import round_sig, fsli_to_xcn
from scipy.special import ndtri as z
from surpyval import nonparametric as nonp
from copy import deepcopy, copy
from matplotlib.ticker import FixedLocator
from autograd import elementwise_grad


CB_COLOUR = "#e94c54"


def _round_vals(x):
    not_different = True
    i = 1
    while not_different:
        x_ticks = np.array(round_sig(x, i))
        not_different = (np.diff(x_ticks) == 0).any()
        i += 1
    return x_ticks


class Parametric():
    """
    Result of ``.fit()`` or ``.from_params()`` method for every parametric
    surpyval distribution.

    Instances of this class are very useful when a user needs the other
    functions of a distribution for plotting, optimizations, monte carlo
    analysis and numeric integration.
    """
    def __init__(self, dist, method, data, offset, lfp, zi):
        self.dist = dist
        self.k = copy(dist.k)
        self.method = method
        self.data = data
        self.offset = offset
        self.lfp = lfp
        self.zi = zi

        bounds = deepcopy(dist.bounds)
        param_map = dist.param_map.copy()

        if offset:
            if data is not None:
                bounds = ((None, np.min(data['x'])), *bounds)
            else:
                bounds = ((None, None), *bounds)

            param_map = {k: v + 1 for k, v in param_map.items()}
            param_map.update({'gamma': 0})
            self.k += 1
        else:
            self.gamma = 0

        if lfp:
            bounds = (*bounds, (0, 1))
            param_map.update({'p': len(param_map) + 1})
            self.k += 1
        else:
            self.p = 1

        if zi:
            bounds = (*bounds, (0, 1))
            param_map.update({'f0': len(param_map) + 1})
            self.k += 1
        else:
            self.f0 = 0

        self.bounds = bounds
        self.param_map = param_map

    def __repr__(self):
        if hasattr(self, 'params'):
            param_string = '\n'.join(['{:>10}'.format(name)
                                      + ": " + str(p) for p, name in
                                      zip(self.params, self.dist.param_names)])
            out = ('Parametric SurPyval Model'
                   + '\n========================='
                   + '\nDistribution        : {dist}'
                   + '\nFitted by           : {method}'
                   ).format(dist=self.dist.name,
                            method=self.method)
            if self.offset:
                out += '\nOffset (gamma)      : {g}'.format(g=self.gamma)

            if self.lfp:
                out += '\nMax Proportion (p)  : {p}'.format(p=self.p)

            if self.zi:
                out += '\nZero-Inflation (f0) : {f0}'.format(f0=self.f0)

            out = (out + '\nParameters          :\n'
                       + '{params}'.format(params=param_string))

            return out
        else:
            return "Unable to fit values"

    def param_cb(self, name, alpha_ci=0.05, bound="two-sided"):
        """
        Method to calculate the confidence bound on a parameter.
        """
        idx = self.dist.param_map[name]
        p_hat = self.params[idx]

        if bound == "two-sided":
            alpha = alpha / 2
            bounds = np.array([-1, 1])
        elif bound == "lower":
            bounds = np.array([-1])
        elif bound == "upper":
            bounds = np.array([1])

        if self.dist.bounds[idx] == (0, None):
            exponent = z(alpha) * np.sqrt(self.hess_inv[idx, idx])/p_hat
            bounds = -bounds * exponent
            return p_hat * np.exp(bounds)
        else:
            factor = z(alpha) * np.sqrt(self.hess_inv[idx, idx])
            bounds = -bounds * factor
            return p_hat + bounds

    def sf(self, x):
        r"""

        Surival (or Reliability) function for a distribution using the
        parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the survival
            function will be calculated.

        Returns
        -------

        sf : scalar or numpy array
            The scalar value of the survival function of the distribution if
            a scalar was passed. If an array like object was passed then a
            numpy array is returned with the value of the survival function at
            each corresponding value in the input array.

        Examples
        --------
        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.sf(2)
        0.9920319148370607
        >>> model.sf([1, 2, 3, 4, 5])
        array([0.9990005 , 0.99203191, 0.97336124, 0.938005  , 0.8824969 ])
        """
        if type(x) == list:
            x = np.array(x)
        if self.p == 1:
            sf = self.dist.sf(x - self.gamma, *self.params)
        else:
            sf = 1 - self.p * self.dist.ff(x - self.gamma, *self.params)
        return (1. - self.f0) * sf

    def ff(self, x):
        r"""

        The cumulative distribution function, or failure function, for a
        distribution using the parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the failure function
            (CDF) will be calculated.

        Returns
        -------

        ff : scalar or numpy array
            The scalar value of the CDF of the distribution if a scalar was
            passed. If an array like object was passed then a numpy array is
            returned with the value of the CDF at each corresponding value in
            the input array.

        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.ff(2)
        0.007968085162939342
        >>> model.ff([1, 2, 3, 4, 5])
        array([0.0009995 , 0.00796809, 0.02663876, 0.061995  , 0.1175031 ])
        """
        if type(x) == list:
            x = np.array(x)

        return self.f0 + ((1 - self.f0)
                          * self.p
                          * self.dist.ff(x - self.gamma, *self.params))

    def df(self, x):
        r"""

        The density function for a distribution using the parameters found
        in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the density function
            will be calculated.

        Returns
        -------

        df : scalar or numpy array
            The scalar value of the density function of the distribution if
            a scalar was passed. If an array like object was passed then a
            numpy array is returned with the value of the density function at
            each corresponding value in the input array.

        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.df(2)
        0.01190438297804473
        >>> model.df([1, 2, 3, 4, 5])
        array([0.002997  , 0.01190438, 0.02628075, 0.04502424, 0.06618727])
        """
        if type(x) == list:
            x = np.array(x)
        if self.f0 == 0:
            df = self.p * self.dist.df(x - self.gamma, *self.params)
        else:
            df = np.where(x == 0, self.f0,
                          ((1 - self.f0)
                           * self.p
                           * self.dist.df(x - self.gamma, *self.params)))
        return df

    def hf(self, x):
        r"""
        The instantaneous hazard function for a distribution using the
        parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the instantaneous
            hazard function will be calculated.

        Returns
        -------

        hf : scalar or numpy array
            The scalar value of the instantaneous hazard function of the
            distribution if a scalar was passed. If an array like object was
            passed then a numpy array is returned with the value of the
            instantaneous hazard function at each corresponding value in
            the input array.

        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.hf(2)
        0.012000000000000002
        >>> model.hf([1, 2, 3, 4, 5])
        array([0.003, 0.012, 0.027, 0.048, 0.075])
        """
        if type(x) == list:
            x = np.array(x)
        if self.p == 1:
            return self.dist.hf(x - self.gamma, *self.params)
        else:
            return 1 - self.p * self.dist.ff(x - self.gamma, *self.params)

    def Hf(self, x):
        r"""
        The cumulative hazard function for a distribution using the
        parameters found in the ``.params`` attribute.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the cumulative
            hazard function will be calculated

        Returns
        -------

        Hf : scalar or numpy array
            The scalar value of the cumulative hazard function of the
            distribution if a scalar was passed. If an array like object was
            passed then a numpy array is returned with the value of the
            cumulative hazard function at each corresponding value in the
            input array.

        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.Hf(2)
        0.008000000000000002
        >>> model.Hf([1, 2, 3, 4, 5])
        array([0.001, 0.008, 0.027, 0.064, 0.125])
        """
        if type(x) == list:
            x = np.array(x)

        if self.p == 1:
            return self.dist.Hf(x - self.gamma, *self.params)
        else:
            return -np.log(self.sf(x))

    def qf(self, p):
        r"""

        The quantile function for a distribution using the parameters found
        in the ``.params`` attribute.

        Parameters
        ----------
        p : array like or scalar
            The values, which must be between 0 and 1, at which the the
            quantile will be calculated

        Returns
        -------
        qf : scalar or numpy array
            The scalar value of the quantile of the distribution if a
            scalar was passed. If an array like object was passed then a
            numpy array is returned with the value of the quantile at each
            corresponding value in the input array.

        Examples
        --------
        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.qf(0.2)
        6.06542793124108
        >>> model.qf([.1, .2, .3, .4, .5])
        array([4.72308719, 6.06542793, 7.09181722, 7.99387877, 8.84997045])
        """
        if type(p) == list:
            p = np.array(p)
        if self.p == 1:
            return self.dist.qf(p, *self.params) + self.gamma
        else:
            raise NotImplementedError("Quantile for LFP not implemented.")

    def cs(self, x, X):
        r"""

        The conditional survival of the model.

        Parameters
        ----------

        x : array like or scalar
            The values at which conditional survival is to be calculated.
        X : array like or scalar
            The value(s) at which it is known the item has survived

        Returns
        -------

        cs : array
            The conditional survival probability.

        Examples
        --------

        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.cs(11, 10)
        0.00025840046151723767
        """
        x = np.array(x)
        cs = np.array(self.dist.cs(x - self.gamma,
                                   X - self.gamma,
                                   *self.params))
        cs[cs > 1.] = 1
        return cs

    def random(self, size, a=None, b=None):
        r"""

        A method to draw random samples from the distributions using the
        parameters found in the ``.params`` attribute.

        Parameters
        ----------
        size : int
            The number of random samples to be drawn from the distribution.
        a: float or None
            The left truncated value if sampling from a truncated
            distribution
        b: float or None
            The right truncated value if sampling from a truncated
            distribution

        Returns
        -------
        random : numpy array
            Returns a numpy array of size ``size`` with random values
            drawn from the distribution.

        Examples
        --------
        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> np.random.seed(1)
        >>> model.random(1)
        array([8.14127103])
        >>> model.random(10)
        array([10.84103403,  0.48542084,  7.11387062,  5.41420125, 4.59286657,
                5.90703589,  7.5124326 ,  7.96575225,  9.18134126, 8.16000438])
        """
        if ((a is not None) | (b is not None)) & ((self.p != 1) |
                                                  (self.f0 != 0)):
            raise NotImplementedError("Truncated smapling not supported with" +
                                      " LFP or ZI models")
        elif ((a is not None) | (b is not None)) & (self.offset):
            raise NotImplementedError("Truncated smapling not supported with" +
                                      " offset distributions")

        if (self.p == 1) and (self.f0 == 0):
            if (a is None) & (b is None):
                return self.dist.qf(uniform.rvs(size=size),
                                    *self.params) + self.gamma

            else:
                # Truncated sampling
                # F-1(u) = G-1[u(G(b) - G(a)) + G(a)]
                if a is None:
                    Fa = 0
                else:
                    Fa = self.dist.ff(a, *self.params)
                if b is None:
                    Fb = 1
                else:
                    Fb = self.dist.ff(b, *self.params)
                u = uniform.rvs(size=size)
                return self.dist.qf((u * (Fb - Fa) + Fa), *self.params)

        elif (self.p != 1) and (self.f0 == 0):
            n_obs = np.random.binomial(size, self.p)

            f = self.dist.qf(uniform.rvs(size=n_obs),
                             *self.params) + self.gamma
            s = np.ones(np.array(size) - n_obs) * np.max(f) + 1

            return fsli_to_xcn(f, s)

        elif (self.p == 1) and (self.f0 != 0):
            n_doa = np.random.binomial(size, self.f0)

            x0 = np.zeros(n_doa) + self.gamma
            x = self.dist.qf(uniform.rvs(size=size - n_doa),
                             *self.params) + self.gamma
            x = np.concatenate([x, x0])
            np.random.shuffle(x)

            return x
        else:
            N = np.random.multinomial(1, [self.f0,
                                          self.p - self.f0,
                                          1. - self.p], size).sum(axis=0)

            N = np.atleast_2d(N)
            n_doa, n_obs, n_cens = N[:, 0], N[:, 1], N[:, 2]
            x0 = np.zeros(n_doa) + self.gamma

            x = self.dist.qf(uniform.rvs(size=n_obs),
                             *self.params) + self.gamma

            f = np.concatenate([x, x0])
            s = np.ones(n_cens) * np.max(f) + 1
            return fsli_to_xcn(f, s)

    def mean(self):
        r"""
        A method to draw random samples from the distributions using the
        parameters found in the ``.params`` attribute.

        Returns
        -------
        mean : float
            Returns the mean of the distribution.

        Examples
        --------
        >>> from surpyval import Weibull
        >>> model = Weibull.from_params([10, 3])
        >>> model.mean()
        8.929795115692489
        """
        if not hasattr(self, '_mean'):
            self._mean = self.p * (self.dist.mean(*self.params) + self.gamma)
        return self._mean

    def moment(self, n):
        r"""

        A method to draw random samples from the distributions using the
        parameters found in the ``.params`` attribute.

        Parameters
        ----------
        n : integer
            The degree of the moment to be computed

        Returns
        -------
        moment[n] : float
            Returns the n-th moment of the distribution

        References
        ----------
        INSERT WIKIPEDIA HERE

        Examples
        --------
        >>> from surpyval import Normal
        >>> model = Normal.from_params([10, 3])
        >>> model.moment(1)
        10.0
        >>> model.moment(5)
        202150.0
        """
        if self.p == 1:
            return self.dist.moment(n, *self.params)
        else:
            msg = 'LFP distributions cannot yet have their moment calculated'
            raise NotImplementedError(msg)

    def entropy(self):
        r"""
        A method to draw random samples from the distributions using the
        parameters found in the ``.params`` attribute.

        Returns
        -------

        entropy : float
            Returns entropy of the distribution

        Examples
        --------
        >>> from surpyval import Normal
        >>> model = Normal.from_params([10, 3])
        >>> model.entropy()
        2.588783247593625
        """
        if self.p == 1:
            return self.dist.entropy(*self.params)
        else:
            msg = "Entropy not available for LFP distribution"
            raise NotImplementedError(msg)

    def cb(self, t, on='R', alpha_ci=0.05, bound='two-sided'):
        r"""
        Confidence bounds of the ``on`` function at the ``alpa_ci`` level of
        significance. Can be the upper, lower, or two-sided confidence by
        changing value of ``bound``.

        Parameters
        ----------

        x : array like or scalar
            The values of the random variables at which the confidence bounds
            will be calculated
        on : ('sf', 'ff', 'Hf'), optional
            The function on which the confidence bound will be calculated.
        bound : ('two-sided', 'upper', 'lower'), str, optional
            Compute either the two-sided, upper or lower confidence bound(s).
            Defaults to two-sided.
        alpha_ci : scalar, optional
            The level of significance at which the bound will be computed.

        Returns
        -------

        cb : scalar or numpy array
            The value(s) of the upper, lower, or both confidence bound(s) of
            the selected function at x

        """
        t = np.atleast_1d(t)
        if self.method != 'MLE':
            raise Exception('Only MLE has confidence bounds')

        hess_inv = np.copy(self.hess_inv)

        pvars = hess_inv[np.triu_indices(hess_inv.shape[0])]
        old_err_state = np.seterr(all='ignore')

        if hasattr(self.dist, 'R_cb'):
            def R_cb(x, bound=bound):
                return self.dist.R_cb(x - self.gamma,
                                      *self.params,
                                      hess_inv,
                                      alpha_ci=alpha_ci,
                                      bound=bound)

        else:
            def R_cb(x, bound=bound):
                def sf_func(params):
                    return self.dist.sf(x - self.gamma, *params)
                jac = np.atleast_2d(jacobian(sf_func)(np.array(self.params)))

                # Second-Order Taylor Series Expansion of Variance
                var_R = []
                for i, j in enumerate(jac):
                    j = np.atleast_2d(j).T * j
                    j = j[np.triu_indices(j.shape[0])]
                    var_R.append(np.sum(j * pvars))

                # First-Order Taylor Series Expansion of Variance
                # var_R = (jac**2 * np.diag(hess_inv)).sum(axis=1).T

                R_hat = self.sf(x)
                if bound == 'two-sided':
                    diff = (z(alpha_ci / 2)
                            * np.sqrt(np.array(var_R))
                            * np.array([1., -1.]).reshape(2, 1))
                elif bound == 'upper':
                    diff = z(alpha_ci) * np.sqrt(np.array(var_R))
                else:
                    diff = -z(alpha_ci) * np.sqrt(np.array(var_R))

                exponent = diff / (R_hat * (1 - R_hat))
                R_cb = R_hat / (R_hat + (1 - R_hat) * np.exp(exponent))
                return R_cb.T

        # Reverse for ff and F
        if on in ['ff', 'F', 'Hf', 'hf', 'df'] and bound == 'lower':
            bound = 'upper'
        elif on in ['ff', 'F', 'Hf', 'hf', 'df'] and bound == 'upper':
            bound = 'lower'

        if (on == 'ff') or (on == 'F'):
            cb = R_cb(t, bound=bound)
            cb = 1. - cb
        elif (on == 'sf') or (on == 'R'):
            cb = R_cb(t, bound=bound)
            if bound == 'two-sided':
                cb = np.fliplr(cb)
        elif on == 'Hf':
            cb = R_cb(t, bound=bound)
            cb = -np.log(cb)
        elif on == 'hf':
            def cb_hf(x):
                out = []
                func = lambda x: -np.log(R_cb(x, bound=bound))
                if bound == 'two-sided':
                    jac = jacobian(func)
                    cbs = [jac(np.array([v])).flatten() for v in x]
                    return np.vstack(cbs)
                else:
                    grad = elementwise_grad(func)
                    return grad(x)

            cb = cb_hf(t)

        elif on == 'df':
            def cb_df(x):
                out = []
                
                func = lambda x: (-np.log(R_cb(x, bound)))
                
                if bound == 'two-sided':
                    jac = jacobian(func)
                    cbs = [jac(np.array([v])).flatten() for v in x]
                    cbs = np.vstack(cbs)
                    cbs = cbs * self.sf(x).reshape(-1, 1)
                else:
                    grad = elementwise_grad(func)
                    cbs = grad(x)
                    cbs = cbs * self.sf(x)

                return cbs

            cb = cb_df(t)

        np.seterr(**old_err_state)
        return cb

    def neg_ll(self):
        r"""

        The the negative log-likelihood for the model, if it was fit with the
        ``fit()`` method. Not available if fit with the ``from_params()``
        method.

        Returns
        -------

        neg_ll : float
            The negative log-likelihood of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.neg_ll()
        262.52685642385734
        """
        if not hasattr(self, 'data'):
            raise ValueError("Must have been fit with data")

        return self._neg_ll

    def bic(self):
        r"""

        The the Bayesian Information Criterion (BIC) for the model, if it
        was fit with the ``fit()`` method. Not available if fit with the
        ``from_params()`` method.

        Returns
        -------

        bic : float
            The BIC of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.bic()
        534.2640532196908

        References:
        -----------

        `Bayesian Information Criterion for Censored Survival Models
        <https://www.jstor.org/stable/2677130>`_.

        """
        if hasattr(self, '_bic'):
            return self._bic
        else:
            self._bic = (self.k
                         * np.log(self.data['n'][self.data['c'] == 0].sum())
                         + 2 * self.neg_ll())
            return self._bic

    def aic(self):
        r"""
        The the Aikake Information Criterion (AIC) for the model, if it was
        fit with the ``fit()`` method. Not available if fit with the
        ``from_params()`` method.

        Returns
        -------

        aic : float
            The AIC of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.aic()
        529.0537128477147
        """
        if hasattr(self, '_aic'):
            return self._aic
        else:
            self._aic = 2 * self.k + 2 * self.neg_ll()
            return self._aic

    def aic_c(self):
        r"""
        The the Corrected Aikake Information Criterion (AIC) for the model,
        if it was fit with the ``fit()`` method. Not available if fit with
        the ``from_params()`` method.

        Returns
        -------

        aic_c : float
            The Corrected AIC of the model

        Examples
        --------

        >>> from surpyval import Weibull
        >>> import numpy as np
        >>> np.random.seed(1)
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.aic()
        529.1774241879209
        """
        if hasattr(self, '_aic_c'):
            return self._aic_c
        else:
            k = len(self.params)
            n = self.data['n'].sum()
            self._aic_c = self.aic() + (2 * k**2 + 2 * k) / (n - k - 1)
            return self._aic_c

    def get_plot_data(self, heuristic='Nelson-Aalen', alpha_ci=0.05):
        r"""

        A method to gather plot data

        Parameters
        ----------
        heuristic : {'Blom', 'Median', 'ECDF', 'Modal', 'Midpoint', 'Mean',
        'Weibull', 'Benard', 'Beard', 'Hazen', 'Gringorten', 'None', 'Tukey',
        'DPW', 'Fleming-Harrington', 'Kaplan-Meier', 'Nelson-Aalen',
        'Filliben', 'Larsen', 'Turnbull'}, optional
            The method that the plotting point on the probablility plot will
            be calculated.

        alpha_ci : float, optional
            The confidence with which the confidence bounds, if able, will
            be calculated. Defaults to 0.95.

        Returns
        -------
        data : dict
            Returns dictionary containing the data needed to do a plot.

        Examples
        --------
        >>> from surpyval import Weibull
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> data = model.get_plot_data()
        """
        x_, r, d, F = nonp.plotting_positions(
            x=self.data['x'],
            c=self.data['c'],
            n=self.data['n'],
            t=self.data['t'],
            heuristic=heuristic)

        mask = np.isfinite(x_)
        x_ = x_[mask] - self.gamma
        r = r[mask]
        d = d[mask]
        F = F[mask]

        if np.isfinite(self.data['t']).any():
            Ftl = self.ff(x_[0])
            Ftr = self.ff(x_[-1])
            F = Ftl + F * (Ftr - Ftl)

        y_scale_min = np.min(F[F > 0]) / 2
        # y_scale_max = np.max(F[F < 1]) + (1 - np.max(F[F < 1]))/2
        y_scale_max = (1 - (1 - np.max(F[F < 1])) / 10)

        # x-axis
        if self.dist.plot_x_scale == 'log':
            log_x = np.log10(x_[x_ > 0])
            x_min = np.min(log_x)
            x_max = np.max(log_x)
            vals_non_sig = 10 ** np.linspace(x_min, x_max, 7)
            x_minor_ticks = np.arange(np.floor(x_min), np.ceil(x_max))
            x_minor_ticks = ((10**x_minor_ticks * np.array(np.arange(1, 11))
                             .reshape((10, 1)))
                             .flatten())
            diff = (x_max - x_min) / 10
            x_scale_min = 10**(x_min - diff)
            x_scale_max = 10**(x_max + diff)
            x_model = 10**np.linspace(x_min - diff, x_max + diff, 100)
        elif self.dist.name in ('Beta'):
            x_min = np.min(x_)
            x_max = np.max(x_)
            x_scale_min = 0
            x_scale_max = 1
            vals_non_sig = np.linspace(x_scale_min, x_scale_max, 11)[1:-1]
            x_minor_ticks = np.linspace(x_scale_min, x_scale_max, 22)[1:-1]
            x_model = np.linspace(x_scale_min, x_scale_max, 102)[1:-1]
        elif self.dist.name in ('Uniform'):
            x_min = np.min(self.params)
            x_max = np.max(self.params)
            x_scale_min = x_min
            x_scale_max = x_max
            vals_non_sig = np.linspace(x_scale_min, x_scale_max, 11)[1:-1]
            x_minor_ticks = np.linspace(x_scale_min, x_scale_max, 22)[1:-1]
            x_model = np.linspace(x_scale_min, x_scale_max, 102)[1:-1]
        else:
            x_min = np.min(x_)
            x_max = np.max(x_)
            vals_non_sig = np.linspace(x_min, x_max, 7)
            x_minor_ticks = np.arange(np.floor(x_min), np.ceil(x_max))
            diff = (x_max - x_min) / 10
            x_scale_min = x_min - diff
            x_scale_max = x_max + diff
            x_model = np.linspace(x_scale_min, x_scale_max, 100)

        cdf = self.ff(x_model + self.gamma)

        x_ticks = _round_vals(vals_non_sig)
        x_ticks_labels = [str(int(x))
                          if (re.match(r'([0-9]+\.0+)', str(x)) is not None)
                          & (x > 1)
                          else str(x)
                          for x in _round_vals(vals_non_sig + self.gamma)]

        y_ticks = np.array(self.dist.y_ticks)
        y_ticks = y_ticks[np.where((y_ticks > y_scale_min)
                          & (y_ticks < y_scale_max))[0]]

        y_ticks_labels = [str(int(y)) + '%'
                          if (re.match(r'([0-9]+\.0+)', str(y)) is not None)
                          & (y > 1)
                          else str(y) for y in y_ticks * 100]

        if hasattr(self, 'hess_inv') & (self.method == 'MLE'):
            if (self.hess_inv is not None):
                cbs = self.cb(x_model + self.gamma, on='ff', alpha_ci=alpha_ci)
            else:
                cbs = []
        else:
            cbs = []

        return {
            'x_scale_min': x_scale_min,
            'x_scale_max': x_scale_max,
            'y_scale_min': y_scale_min,
            'y_scale_max': y_scale_max,
            'y_ticks': y_ticks,
            'y_ticks_labels': y_ticks_labels,
            'x_ticks': x_ticks,
            'x_ticks_labels': x_ticks_labels,
            'cdf': cdf,
            'x_model': x_model,
            'x_minor_ticks': x_minor_ticks,
            'cbs': cbs,
            'x_scale': self.dist.plot_x_scale,
            'x_': x_,
            'F': F
        }

    def plot(self, heuristic='Nelson-Aalen',
             plot_bounds=True,
             alpha_ci=0.05,
             ax=None):
        r"""
        A method to do a probability plot

        Parameters
        ----------
        heuristic : {'Blom', 'Median', 'ECDF', 'Modal', 'Midpoint', 'Mean',
        'Weibull', 'Benard', 'Beard', 'Hazen', 'Gringorten', 'None', 'Tukey',
        'DPW', 'Fleming-Harrington', 'Kaplan-Meier', 'Nelson-Aalen',
        'Filliben', 'Larsen', 'Turnbull'}, optional
            The method that the plotting point on the probablility plot will
            be calculated.

        plot_bounds : Boolean, optional
            A Boolean value to indicate whehter you want the probability
            bounds to be calculated.

        alpha_ci : float, optional
            The confidence with which the confidence bounds, if able, will
            be calculated. Defaults to 0.95.

        ax: matplotlib.axes.Axes, optional
            The axis onto which the plot will be created. Optional, if not
            provided a new axes will be created.

        Returns
        -------
        plot : list
            list of a matplotlib plot object

        Examples
        --------
        >>> from surpyval import Weibull
        >>> x = Weibull.random(100, 10, 3)
        >>> model = Weibull.fit(x)
        >>> model.plot()
        """
        if ax is None:
            ax = plt.gcf().gca()

        if not hasattr(self, 'params'):
            raise Exception("Can't plot model that failed to fit")

        if self.method == 'given parameters':
            detail = "Can't plot model that was given parameters and no data"
            raise Exception(detail)

        d = self.get_plot_data(heuristic=heuristic, alpha_ci=alpha_ci)

        # Set limits and scale
        ax.set_ylim([d['y_scale_min'], d['y_scale_max']])
        ax.set_xscale(d['x_scale'])
        functions = (
            lambda x: self.dist.mpp_y_transform(x, *self.params),
            lambda x: self.dist.mpp_inv_y_transform(x, *self.params))
        ax.set_yscale('function', functions=functions)
        ax.set_yticks(d['y_ticks'])
        ax.set_yticklabels(d['y_ticks_labels'])
        ax.yaxis.set_minor_locator(FixedLocator(np.linspace(0, 1, 51)))
        ax.set_xticks(d['x_ticks'])
        ax.set_xticklabels(d['x_ticks_labels'])

        if d['x_scale'] == 'log':
            ax.set_xticks(d['x_minor_ticks'], minor=True)
            ax.set_xticklabels([], minor=True)

        ax.grid(b=True, which='major', color='g', alpha=0.4, linestyle='-')
        ax.grid(b=True, which='minor', color='g', alpha=0.1, linestyle='-')

        ax.set_title('{} Probability Plot'.format(self.dist.name))
        ax.set_ylabel('CDF')
        ax.scatter(d['x_'], d['F'])

        ax.set_xlim([d['x_scale_min'], d['x_scale_max']])
        if plot_bounds & (len(d['cbs']) != 0):
            ax.plot(d['x_model'], d['cbs'], color=CB_COLOUR)

        ax.plot(d['x_model'], d['cdf'], color='k', linestyle='--')
        return ax



from surpyval import np
from scipy.stats import uniform
from numpy import euler_gamma
from scipy.special import gamma as gamma_func
from scipy.special import ndtri as z

from surpyval import parametric as para
from surpyval.parametric.parametric_fitter import ParametricFitter

from .fitters.mpp import mpp

class Weibull_(ParametricFitter):
    def __init__(self, name):
        self.name = name
        # Set 'k', the number of parameters
        self.k = 2
        self.bounds = ((0, None), (0, None),)
        self.support = (0, np.inf)
        self.plot_x_scale = 'log'
        self.y_ticks = [0.0001, 0.0002, 0.0003, 0.001, 0.002, 
            0.003, 0.005, 0.01, 0.02, 0.03, 0.05, 
            0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 
            0.9, 0.95, 0.99, 0.999, 0.9999]
        self.param_names = ['alpha', 'beta']
        self.param_map = {
            'alpha' : 0,
            'beta' : 1
        }

    def _parameter_initialiser(self, x, c=None, n=None, t=None, offset=False):
        log_x = np.log(x)
        log_x[np.isnan(log_x)] = -np.inf
        if (2 in c) or (-1 in c):
            heuristic = "Turnbull"
        else:
            heuristic = "Fleming-Harrington"

        data = {'x' : x, 'c' : c, 'n' : n, 't' : t}
        model = para.Parametric(self, 'MPP', data, offset, False, False)
        fitting_info = {}
        fitting_info['rr'] = 'x'
        fitting_info['heuristic'] = heuristic
        fitting_info['on_d_is_0'] = True
        fitting_info['turnbull_estimator'] = 'Fleming-Harrington'
        fitting_info['init'] = None

        model.fitting_info = fitting_info

        if offset:
            results = mpp(model)
            return (results['gamma'], *results['params'])
        else:
            gumb = para.Gumbel.fit(log_x, c, n, t, how='MLE')
            if not gumb.res.success:
                gumb = para.Gumbel.fit(log_x, c, n, t, how='MPP', heuristic=heuristic)
            mu, sigma = gumb.params
            alpha, beta = np.exp(mu), 1. / sigma
            if (np.isinf(alpha) | np.isnan(alpha)):
                alpha = np.median(x)
            if (np.isinf(beta) | np.isnan(beta)):
                beta = 1.
            return alpha, beta

    def sf(self, x, alpha, beta):
        r"""

        Survival (or reliability) function for the Weibull Distribution:

        .. math::
            R(x) = e^{-\left ( \frac{x}{\alpha} \right )^\beta}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the reliability function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Weibull.sf(x, 3, 4)
        array([9.87730216e-01, 8.20754808e-01, 3.67879441e-01, 4.24047953e-02,
               4.45617596e-04])
        """
        return np.exp(-(x / alpha)**beta)

    def ff(self, x, alpha, beta):
        r"""

        Failure (CDF or unreliability) function for the Weibull Distribution:

        .. math::
            F(x) = 1 - e^{-\left ( \frac{x}{\alpha} \right )^\beta}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        sf : scalar or numpy array 
            The value(s) of the failure function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Weibull.ff(x, 3, 4)
        array([0.01226978, 0.17924519, 0.63212056, 0.9575952 , 0.99955438])
        """
        return 1 - np.exp(-(x / alpha)**beta)

    def cs(self, x, X, alpha, beta):
        r"""

        Conditional survival function for the Weibull Distribution:

        .. math::
            R(x, X) = \frac{R(x + X)}{R(X)}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        cs : scalar or numpy array 
            The value(s) of the conditional survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Weibull.cs(x, 5, 3, 4)
        array([2.21654222e+03, 1.84183662e+03, 8.25549630e+02, 9.51596070e+01,
               1.00000000e+00])
        """
        return self.sf(x, alpha, beta) / self.sf(X, alpha, beta)

    def df(self, x, alpha, beta):
        r"""

        Density function for the Weibull Distribution:

        .. math::
            f(x) = \frac{\beta}{\alpha} \frac{x}{\alpha}^{\beta - 1} e^{-\left ( \frac{x}{\alpha} \right )^\beta}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the conditional survival function at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Weibull.df(x, 5, 3, 4)
        array([0.0487768 , 0.32424881, 0.49050592, 0.13402009, 0.00275073])
        """
        return (beta / alpha) * (x / alpha)**(beta-1) * np.exp(-(x / alpha)**beta)

    def hf(self, x, alpha, beta):
        r"""

        Instantaneous hazard rate for the Weibull Distribution:

        .. math::
            h(x) = \frac{\beta}{\alpha} \left ( \frac{x}{\alpha} \right )^{\beta - 1}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the instantaneous hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Weibull.hf(x, 3, 4)
        array([0.04938272, 0.39506173, 1.33333333, 3.16049383, 6.17283951])
        """
        return (beta / alpha) * (x / alpha)**(beta - 1)

    def Hf(self, x, alpha, beta):
        r"""

        Cumulative hazard rate for the Weibull Distribution:

        .. math::
            h(x) = \frac{x}{\alpha}^{\beta}

        Parameters
        ----------

        x : numpy array or scalar
            The values at which the function will be calculated 
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        df : scalar or numpy array 
            The value(s) of the cumulative hazard rate at x.

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> Weibull.Hf(x, 3, 4)
        array([0.01234568, 0.19753086, 1.        , 3.16049383, 7.71604938])
        """
        return (x / alpha)**beta

    def qf(self, p, alpha, beta):
        r"""

        Quantile function for the Weibull distribution:

        .. math::
            q(p) = \alpha \left ( -\ln \left ( 1 - p \right ) \right )^{1/ \beta}

        Parameters
        ----------

        p : numpy array or scalar
            The percentiles at which the quantile will be calculated
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        q : scalar or numpy array 
            The quantiles for the Weibull distribution at each value p

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> p = np.array([.1, .2, .3, .4, .5])
        >>> Weibull.qf(p, 3, 4)
        array([1.70919151, 2.06189877, 2.31840554, 2.5362346 , 2.73733292])
        """
        return alpha * (-np.log(1 - p))**(1/beta)

    def mean(self, alpha, beta):
        r"""

        Mean of the Weibull distribution

        .. math::
            E = \alpha \Gamma \left ( 1 + \frac{1}{\beta} \right )

        Parameters
        ----------

        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        mean : scalar or numpy array 
            The mean(s) of the Weibull distribution

        Examples
        --------
        >>> from surpyval import Weibull
        >>> Weibull.mean(3, 4)
        2.7192074311664314
        """
        return alpha * gamma_func(1 + 1./beta)

    def moment(self, n, alpha, beta):
        r"""

        n-th moment of the Weibull distribution

        .. math::
            M(n) = \alpha^n \Gamma \left ( 1 + \frac{n}{\beta} \right )

        Parameters
        ----------

        n : integer or numpy array of integers
            The ordinal of the moment to calculate
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        mean : scalar or numpy array 
            The moment(s) of the Weibull distribution

        Examples
        --------
        >>> from surpyval import Weibull
        >>> Weibull.moment(2, 3, 4)
        7.976042329074821
        """
        return alpha**n * gamma_func(1 + n/beta)

    def entropy(self, alpha, beta):
        return euler_gamma * (1 - 1/beta) + np.log(alpha / beta) + 1

    def random(self, size, alpha, beta):
        r"""

        Draws random samples from the distribution in shape `size`

        Parameters
        ----------

        size : integer or tuple of positive integers
            Shape or size of the random draw
        alpha : numpy array or scalar
            scale parameter for the Weibull distribution
        beta : numpy array or scalar
            shape parameter for the Weibull distribution

        Returns
        -------

        random : scalar or numpy array 
            Random values drawn from the distribution in shape `size`

        Examples
        --------
        >>> import numpy as np
        >>> from surpyval import Weibull
        >>> Weibull.random(10, 3, 4)
        array([1.79782451, 1.7143211 , 2.84778674, 3.12226231, 2.61000839,
               3.05456332, 3.00280851, 2.61910071, 1.37991527, 4.17488394])
        >>> Weibull.random((5, 5), 3, 4)
        array([[1.64782514, 2.79157632, 1.85500681, 2.91908736, 2.46089933],
               [1.85880127, 0.96787742, 2.29677031, 2.42394129, 2.63889601],
               [2.14351859, 3.90677225, 2.24013855, 2.49467774, 3.43755278],
               [3.24417396, 1.40775181, 2.49584969, 3.07603353, 2.54679499],
               [1.98330076, 2.95002633, 3.35402601, 3.11429283, 3.45706789]])
        """
        U = uniform.rvs(size=size)
        return self.qf(U, alpha, beta)

    def mpp_x_transform(self, x):
        return np.log(x)

    def mpp_inv_x_transform(self, x, gamma=0):
        return np.exp(x - gamma)

    def mpp_y_transform(self, y, *params):
        mask = ((y == 0) | (y == 1))
        out = np.zeros_like(y)
        out[~mask] = np.log(-np.log((1 - y[~mask])))
        out[mask] = np.nan
        return out

    def mpp_inv_y_transform(self, y, *params):
        return 1 - np.exp(-np.exp(y))

    def unpack_rr(self, params, rr):
        if rr == 'y':
            beta  = params[0]
            alpha = np.exp(params[1]/-beta)
        elif rr == 'x':
            beta  = 1./params[0]
            alpha = np.exp(params[1] / (beta * params[0]))
        return alpha, beta

    def u(self, x, alpha, beta):
        return beta * (np.log(x) - np.log(alpha))

    def u_cb(self, x, alpha, beta, cv_matrix, alpha_ci, bound='two-sided'):
        u = self.u(x, alpha, beta)
        var_u = self.var_u(x, alpha, beta, cv_matrix)
        if bound == 'two-sided':
            diff = z(alpha_ci/2) * np.array([1., -1.]).reshape(2, 1)
        elif bound == 'upper':
            diff = z(alpha_ci)
        else:
            diff = -z(alpha_ci)
        u_cb = u + (diff * np.sqrt(var_u))
        return u_cb

    def du(self, x, alpha, beta):
        du_dbeta = np.log(x) - np.log(alpha)
        du_dalpha  = -beta/alpha
        return du_dalpha, du_dbeta

    def var_u(self, x, alpha, beta, cv_matrix):
        da, db = self.du(x, alpha, beta)
        var_u = (da**2 * cv_matrix[0, 0] + db**2 * cv_matrix[1, 1] + 
            2 * da * db * cv_matrix[0, 1])
        return var_u

    def R_cb(self, x, alpha, beta, cv_matrix, alpha_ci=0.05, bound='two-sided'):
        return np.exp(-np.exp(self.u_cb(x, alpha, beta, cv_matrix, alpha_ci, bound))).T

    def _jacobian(self, x, alpha, beta, c=None, n=None):
        f = c == 0
        l = c == -1
        r = c == 1
        dll_dbeta = (
            1./beta * np.sum(n[f]) +
            np.sum(n[f] * np.log(x[f]/alpha)) - 
            np.sum(n[f] * (x[f]/alpha)**beta * np.log(x[f]/alpha)) - 
            np.sum(n[r] * (x[r]/alpha)**beta * np.log(x[r]/alpha)) +
            np.sum(n[l] * (x[l]/alpha)**beta * np.log(x[l]/alpha) *
                np.exp(-(x[l]/alpha)**beta) / 
                (1 - np.exp(-(x[l]/alpha)**beta)))
        )

        dll_dalpha = ( 0 -
            beta/alpha * np.sum(n[f]) +
            beta/alpha * np.sum(n[f] * (x[f]/alpha)**beta) +
            beta/alpha * np.sum(n[r] * (x[r]/alpha)**beta) -
            beta/alpha * np.sum(n[l] * (x[l]/alpha)**beta * 
                np.exp(-(x[l]/alpha)**beta) /
                (1 - np.exp(-(x[l]/alpha)**beta)))
        )
        return -np.array([dll_dalpha, dll_dbeta])

Weibull = Weibull_('Weibull')


from surpyval import xcnt_handler
from surpyval import nonparametric as nonp
from pandas import Series
import numpy as np
from surpyval.utils import xcnt_to_xrd


def plotting_positions(x, c=None, n=None, t=None, heuristic="Blom",
                       turnbull_estimator='Fleming-Harrington'):
    r"""
    This function takes in data in the xcnt format and outputs an approximation
    of the CDF. This function can be used to produce estimates of F using the
    Nelson-Aalen, Kaplan-Meier, Fleming-Harrington, and the Turnbull estimates.
    Additionally, it can be used to create 'plotting heuristics.'

    Plotting heuristics are the values that are used to plot on probability
    paper and can be used to estiamte the parameters of a distribution. The use
    of probability plots is one of the traditional ways to estimate the
    parameters of a distribution.

    If right censored data can be used by the regular plotting positions. If
    there is right censored data this method adjusts the ranks of the values
    using the mean order number.

    Parameters
    ----------

    x : array like, optional
        Array of observations of the random variables. If x is :code:`None`,
        xl and xr must be provided.
    c : array like, optional
        Array of censoring flag. -1 is left censored, 0 is observed, 1 is
        right censored, and 2 is intervally censored. If not provided will
        assume all values are observed.
    n : array like, optional
        Array of counts for each x. If data is proivded as counts, then this
        can be provided. If :code:`None` will assume each observation is 1.
    t : 2D-array like, optional
        2D array like of the left and right values at which the respective
        observation was truncated. If not provided it assumes that no
        truncation occurs.
    heuristic : ("Blom", "Median", "ECDF", "ECDF_Adj", "Modal", "Midpoint",
    "Mean", "Weibull", "Benard", "Beard", "Hazen", "Gringorten", "None",
    "Larsen", "Tukey", "DPW"), str, optional
        Method to use to compute the heuristic of F. See details of each
        heursitic in the probability plotting section.
    turnbull_estimator : ('Nelson-Aalen', 'Kaplan-Meier'), str, optional
        If using the Turnbull heuristic, you can elect to use the NA or KM
        method to compute R with the Turnbull estimates of the risk and death
        sets.

    Returns
    -------

    x : numpy array
        x values for the plotting points
    r : numpy array
        risk set at each x
    d : numpy array
        death set at each x
    F : numpy array
        estimate of F to use in plotting positions.

    Examples
    --------

    >>> from surpyval.nonparametric import plotting_positions
    >>> import numpy as np
    >>> x = np.array([1, 2, 3, 4, 5, 6, 7, 8])
    >>> x, r, d, F = plotting_positions(x, heuristic="Filliben")
    >>> F
    array([0.08299596, 0.20113568, 0.32068141, 0.44022714, 0.55977286,
           0.67931859, 0.79886432, 0.91700404])
    """

    x, c, n, t = xcnt_handler(x, c, n, t)

    if heuristic not in nonp.PLOTTING_METHODS:
        raise ValueError("Must use available heuristic")

    if ((-1 in c) or (2 in c)) & (heuristic != 'Turnbull'):
        raise ValueError("Left or interval censored data requires "
                         + "the use of the Turnbull estimator")

    if (np.isfinite(t[:, 0]).any()) & (heuristic not in ['Nelson-Aalen',
                                                         'Kaplan-Meier',
                                                         'Fleming-Harrington',
                                                         'Turnbull']):
        raise ValueError("Left truncated data can only be used with "
                         + "'Nelson-Aalen', 'Kaplan-Meier', "
                         + "'Fleming-Harrington', and 'Turnbull' estimators")

    if (np.isfinite(t[:, 1]).any()) & (heuristic != 'Turnbull'):
        raise ValueError("Right truncated data can only be used with "
                         + "'Turnbull' estimator")

    N = n.sum()

    if heuristic == 'Filliben':
        # Needs work
        out = nonp.filliben(x, c, n)
    elif heuristic == 'Nelson-Aalen':
        x, r, d = xcnt_to_xrd(x, c, n, t)
        R = nonp.nelson_aalen(r, d)
        return x, r, d, 1 - R
    elif heuristic == 'Kaplan-Meier':
        x, r, d = xcnt_to_xrd(x, c, n, t)
        R = nonp.kaplan_meier(r, d)
        return x, r, d, 1 - R
    elif heuristic == 'Fleming-Harrington':
        x, r, d = xcnt_to_xrd(x, c, n, t)
        R = nonp.fleming_harrington(r, d)
        return x, r, d, 1 - R
    elif heuristic == 'Turnbull':
        out = nonp.turnbull(x, c, n, t, estimator=turnbull_estimator)
    else:
        # Reformat for plotting point style
        x_ = np.repeat(x, n)
        c = np.repeat(c, n)
        n = np.ones_like(x_)

        idx = np.argsort(c, kind='stable')
        x_ = x_[idx]
        c = c[idx]

        idx2 = np.argsort(x_, kind='stable')
        x_ = x_[idx2]
        c = c[idx2]

        ranks = nonp.rank_adjust(x_, c=c)
        d = 1 - c
        r = np.linspace(N, 1, num=N)

        if heuristic == "Blom":
            A, B = 0.375, 0.25
        elif heuristic == "Median":
            A, B = 0.3, 0.4
        elif heuristic == "ECDF":
            A, B = 0, 0
        elif heuristic == "ECDF_Adj":
            A, B = 0, 1
        elif heuristic == "Modal":
            A, B = 1.0, -1.0
        elif heuristic == "Midpoint":
            A, B = 0.5, 0.0
        elif heuristic == "Mean":
            A, B = 0.0, 1.0
        elif heuristic == "Weibull":
            A, B = 0.0, 1.0
        elif heuristic == "Benard":
            A, B = 0.3, 0.2
        elif heuristic == "Beard":
            A, B = 0.31, 0.38
        elif heuristic == "Hazen":
            A, B = 0.5, 0.0
        elif heuristic == "Gringorten":
            A, B = 0.44, 0.12
        elif heuristic == "None":
            A, B = 0.0, 0.0
        elif heuristic == "Larsen":
            A, B = 0.567, -0.134
        elif heuristic == "Tukey":
            A, B = 1. / 3., 1. / 3.
        elif heuristic == "DPW":
            A, B = 1.0, 0.0

        F = (ranks - A) / (N + B)
        R = 1 - Series(F).ffill().fillna(0).values
        out = {}
        out['x'] = x_
        out['r'] = r
        out['d'] = d
        out['R'] = R

    return out['x'], out['r'], out['d'], 1 - out['R']
